/* *****************************************************************
 *
 *  $Id: f0m_back_end.c,v 1.4 2002/08/12 00:29:34 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/10 12:18:42 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation of the back-end of the services
 *				 provided by FIDDLE-1.
 *
 *		f0m_be_initialize ()	-- Initialize FIDDLE-1
 *		f0m_be_terminate ()		-- Terminate FIDDLE-1
 *		f0m_be_tids ()			-- List existing task IDs
 *		f0m_be_kill ()			-- Kill a target process
 *		f0m_be_attach ()		-- Attach to a running process
 *		f0m_be_detach ()		-- Detach from a running process
 *		f0m_be_symbol_file ()	-- Specify whre to find program symbols
 *		f0m_be_file ()			-- Load program and symbols into memory
 *		f0m_be_run ()			-- Run a program
 *		f0m_be_step ()			-- Executite until the next instruction
 *		f0m_be_next ()			-- Executite until the next instruction
 *		f0m_be_continue ()		-- Continue the execution
 *		f0m_be_finish ()		-- Execute until returning from
 *								   current stack frame
 *      f0m_be_signal ()        -- Send a signal to a process
 *		f0m_be_break ()			-- Set a breakpoint
 *		f0m_be_delete ()		-- Delete a breakpoint
 *		f0m_be_set_variable ()	-- Change the contents of a variable
 *		f0m_be_call ()			-- Call a function in the current
 *								   context of the target process
 *		f0m_be_evaluate ()		-- Evaluate an expression in the current
 *								   context of the target process
 *		f0m_be_display ()		-- Evaluate and display  an expression 
 *								   every time the process execution stops
 *		f0m_be_undisplay ()		-- Undisplay an expression
 *		f0m_be_list ()			-- List the process source code
 *		f0m_be_info_locals ()	-- Get name and value of local variables
 *		f0m_be_info_args ()		-- Get name and value of function argument
 *		f0m_be_info_line ()		-- Get info about current line
 *		f0m_be_info_program ()	-- Get info about current program
 *		f0m_be_info_stack ()	-- Get info about current stack frames
 *		f0m_be_up ()			-- Go up in the stack frame list
 *		f0m_be_down ()			-- Go down in the stack frame list
 *		f0m_be_frame ()			-- Select a specific stack frame
 *		f0m_be_info_break ()	-- Get info about current breakpoint
 *		f0m_be_info_display ()	-- Get info about current display expressions
 *		f0m_be_thread ()		-- Select a thread
 *		f0m_be_info_threads ()	-- Get info about the existing threads
 *		f0m_be_tty ()			-- Set a TTY for future IO
 *		f0m_be_sendto ()		-- Send a command directly to the debugger
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include <pthread.h>
#include "chain.h"
#include "f0s_lib.h"
#include "f0m_lib.h"
#include "f0m_queues.h"
#include "f0m_tkin.h"
#include "f0m_back_end.h"
#include "config.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#define NO_SERVER -1

typedef tkout_t *(*serv_f)(int tid, const tkin_t *tk);


typedef struct serv_vect_s serv_vect_t;
struct serv_vect_s {
	int     code;
	serv_f	funct;
};


PRIVATE tkout_t *serv_tids (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_kill (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_add_debugger (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_attach (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_detach (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_symbol_file (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_file (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_run (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_step (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_next (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_continue (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_finish (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_signal (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_break (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_delete (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_set_variable (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_evaluate (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_display (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_undisplay (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_call (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_list (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_tty (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_up (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_down (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_frame (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_locals (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_args (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_line (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_program (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_stack (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_break (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_display (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_thread (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_info_threads (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_sendto (int tid, const tkin_t *tk);
PRIVATE tkout_t *serv_terminate_thread (int tid, const tkin_t *tk);




typedef struct thread_info_s thread_info_t;
struct thread_info_s {
	int       tid;
	pthread_t thid;
};

PRIVATE chain_t *serv_q = NULL;	/* of thread_info_t */


PRIVATE serv_vect_t serv_vect[] =
{
	{C_TIDS,			serv_tids},
	{C_KILL,			serv_kill},
	{C_ADD_DEBUGGER,	serv_add_debugger},
	{C_ATTACH,			serv_attach},
	{C_DETACH,			serv_detach},
	{C_SYMBOL_FILE,	serv_symbol_file},
	{C_FILE,			serv_file},
	{C_RUN,		    serv_run},
	{C_STEP,			serv_step},
	{C_NEXT,			serv_next},
	{C_CONTINUE,		serv_continue},
	{C_FINISH,			serv_finish},
	{C_SIGNAL,			serv_signal},
	{C_BREAK,			serv_break},
	{C_DELETE,			serv_delete},
	{C_SET_VARIABLE,	serv_set_variable},
	{C_EVALUATE,		serv_evaluate},
	{C_DISPLAY,		serv_display},
	{C_UNDISPLAY,		serv_undisplay},
	{C_CALL,			serv_call},
	{C_LIST,			serv_list},
	{C_TTY,		    serv_tty},
	{C_UP,				serv_up},
	{C_DOWN,			serv_down},
	{C_FRAME,			serv_frame},
	{C_INFO_LOCALS,		serv_info_locals},
	{C_INFO_ARGS,		serv_info_args},
	{C_INFO_LINE,		serv_info_line},
	{C_INFO_PROGRAM,	serv_info_program},
	{C_INFO_STACK,		serv_info_stack},
	{C_INFO_BREAK,		serv_info_break},
	{C_INFO_DISPLAY,	serv_info_display},
	{C_THREAD,			serv_thread},
	{C_INFO_THREADS,	serv_info_threads},
	{C_SENDTO,			serv_sendto},
	{C_TERM_SERVER,		serv_terminate_thread},
	{C_LAST,			NULL},
};


PRIVATE int f0m__start_service_thread (int tid);


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

PRIVATE tkout_t *serv_tids (int tid, const tkin_t *tk)
{
	assert (tk->code == C_TIDS);
	return f0s_tids (tk->data.r_tids_clients.options);
}

PRIVATE tkout_t *serv_kill (int tid, const tkin_t *tk)
{
	assert (tk->code == C_KILL);
	return f0s_kill (tid);
}

PRIVATE tkout_t *serv_add_debugger (int tid, const tkin_t *tk)
{
	tkout_t *rep;
	assert (tk->code == C_ADD_DEBUGGER);
	rep = f0s_add_debugger (tk->data.r_add_debugger.dbg_pid,
							tk->data.r_add_debugger.tp_pid,
							tk->data.r_add_debugger.cmd_channel,
							tk->data.r_add_debugger.rep_channel);
	if (rep != NULL  &&  rep->status == E_RTS_OK  &&  rep->err == NULL) {
		radddebugger_t *info = (radddebugger_t *)chain_first_data (rep->rep->data);
		assert (info->tid > 0);
		f0m__start_service_thread (info->tid);
	}
	return rep;
}

PRIVATE tkout_t *serv_attach (int tid, const tkin_t *tk)
{
	tkout_t *rep;
	assert (tk->code == C_ATTACH);
	rep = f0s_attach (tk->data.r_attach.tp_pid);
	if (rep != NULL  &&  rep->status == E_RTS_OK  &&  rep->err == NULL) {
		rattach_t *att = (rattach_t *)chain_first_data (rep->rep->data);
		assert (att->tp_pid = tk->data.r_attach.tp_pid);
		assert (att->tid > 0);
		f0m__start_service_thread (att->tid);
	}
	return rep;
}

PRIVATE tkout_t *serv_detach (int tid, const tkin_t *tk)
{
	assert (tk->code == C_DETACH);
	return f0s_detach (tid);
}

PRIVATE tkout_t *serv_symbol_file (int tid, const tkin_t *tk)
{
	assert (tk->code == C_SYMBOL_FILE);
	return f0s_symbol_file (tid, tk->data.r_symbol_file.file);
}

PRIVATE tkout_t *serv_file (int tid, const tkin_t *tk)
{
	tkout_t *rep;
	
	assert (tk->code == C_FILE);
	rep = f0s_file (tid, tk->data.r_file.file);
	if (rep != NULL  &&  rep->status == E_RTS_OK  &&  rep->err == NULL) {
		rfile_t *file = (rfile_t *)chain_first_data (rep->rep->data);
		assert (file->tid > 0);
		f0m__start_service_thread (file->tid);
	}
	return rep;
}

PRIVATE tkout_t *serv_run (int tid, const tkin_t *tk)
{
	assert (tk->code == C_RUN);
	return f0s_run (tid, tk->data.r_run.args);
}

PRIVATE tkout_t *serv_step (int tid, const tkin_t *tk)
{
	assert (tk->code == C_STEP);
	return f0s_step (tid);
}

PRIVATE tkout_t *serv_next (int tid, const tkin_t *tk)
{
	assert (tk->code == C_NEXT);
	return f0s_next (tid);
}

PRIVATE tkout_t *serv_continue (int tid, const tkin_t *tk)
{
	assert (tk->code == C_CONTINUE);
	return f0s_continue (tid);
}

PRIVATE tkout_t *serv_finish (int tid, const tkin_t *tk)
{
	assert (tk->code == C_FINISH);
	return f0s_finish (tid);
}

PRIVATE tkout_t *serv_signal (int tid, const tkin_t *tk)
{
	assert (tk->code == C_SIGNAL);
	return f0s_signal (tk->data.r_signal.tid, tk->data.r_signal.signum);
}

PRIVATE tkout_t *serv_break (int tid, const tkin_t *tk)
{
	assert (tk->code == C_BREAK);
	return f0s_break (tid, tk->data.r_break.file,
					 tk->data.r_break.line,
					 tk->data.r_break.function);
}

PRIVATE tkout_t *serv_delete (int tid, const tkin_t *tk)
{
	assert (tk->code == C_DELETE);
	return f0s_delete (tid, tk->data.r_delete.bpid);
}

PRIVATE tkout_t *serv_set_variable (int tid, const tkin_t *tk)
{
	assert (tk->code == C_SET_VARIABLE);
	return f0s_set_variable (tid, tk->data.r_set_variable.variable,
							tk->data.r_set_variable.expression);
}

PRIVATE tkout_t *serv_evaluate (int tid, const tkin_t *tk)
{
	assert (tk->code == C_EVALUATE);
	return f0s_evaluate (tid, tk->data.r_evaluate.expression);
}

PRIVATE tkout_t *serv_display (int tid, const tkin_t *tk)
{
	assert (tk->code == C_DISPLAY);
	return f0s_display (tid, tk->data.r_display.expression);
}

PRIVATE tkout_t *serv_undisplay (int tid, const tkin_t *tk)
{
	assert (tk->code == C_UNDISPLAY);
	return f0s_undisplay (tid, tk->data.r_undisplay.dispid);
}

PRIVATE tkout_t *serv_call (int tid, const tkin_t *tk)
{
	assert (tk->code == C_CALL);
	return f0s_call (tid, tk->data.r_call.expression);
}

PRIVATE tkout_t *serv_list (int tid, const tkin_t *tk)
{
	assert (tk->code == C_LIST);
	return f0s_list (tid, tk->data.r_list.file,
					tk->data.r_list.start_line,
					tk->data.r_list.function,
					tk->data.r_list.n_lines);
}

PRIVATE tkout_t *serv_tty (int tid, const tkin_t *tk)
{
	assert (tk->code == C_TTY);
	return f0s_tty (tid, tk->data.r_tty.file);
}

PRIVATE tkout_t *serv_up (int tid, const tkin_t *tk)
{
	assert (tk->code == C_UP);
	return f0s_up (tid);
}

PRIVATE tkout_t *serv_down (int tid, const tkin_t *tk)
{
	assert (tk->code == C_DOWN);
	return f0s_down (tid);
}

PRIVATE tkout_t *serv_frame (int tid, const tkin_t *tk)
{
	assert (tk->code == C_FRAME);
	return f0s_frame (tid, tk->data.r_frame.frame);
}

PRIVATE tkout_t *serv_info_locals (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_LOCALS);
	return f0s_info_locals (tid);
}

PRIVATE tkout_t *serv_info_args (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_ARGS);
	return f0s_info_args (tid);
}

PRIVATE tkout_t *serv_info_line (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_LINE);
	return f0s_info_line (tid, tk->data.r_info_line.file,
						 tk->data.r_info_line.line,
						 tk->data.r_info_line.function);
}

PRIVATE tkout_t *serv_info_program (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_PROGRAM);
	return f0s_info_program (tid);
}

PRIVATE tkout_t *serv_info_stack (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_STACK);
	return f0s_info_stack (tid);
}

PRIVATE tkout_t *serv_info_break (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_BREAK);
	return f0s_info_break (tid);
}

PRIVATE tkout_t *serv_info_display (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_DISPLAY);
	return f0s_info_display (tid);
}

PRIVATE tkout_t *serv_thread (int tid, const tkin_t *tk)
{
	assert (tk->code == C_THREAD);
	return f0s_thread (tid, tk->data.r_thread.threadID);
}

PRIVATE tkout_t *serv_info_threads (int tid, const tkin_t *tk)
{
	assert (tk->code == C_INFO_THREADS);
	return f0s_info_threads (tid);
}

PRIVATE tkout_t *serv_sendto (int tid, const tkin_t *tk)
{
	assert (tk->code == C_SENDTO);
	return f0s_sendto (tid, tk->data.r_sendto.command);
}

PRIVATE tkout_t *serv_terminate_thread (int tid, const tkin_t *tk)
{
	assert (tk->code == C_TERM_SERVER);
	pthread_exit (NULL);
	return NULL;
}



PRIVATE int f0m__terminate_thread (void *obj, void *arg)
{
	int reqid;
	tkin_t *req;
	thread_info_t *info = (thread_info_t *)obj;
	if (info->tid != NO_SERVER) {
		req = f0m_tkin_new (info->tid, C_TERM_SERVER, 0);
		reqid = f0m_enqueue_request (info->tid, req);
		assert (reqid >= 0);
	}
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    PRIVATE tkout_t *service(unsigned long tid, tkin_t *tkin)
 *
 * Description: Look in table for appropriate function to process
 *				service 'code'.
 *
 * Returns:     NULL
 */
PRIVATE tkout_t *service(int tid, const tkin_t *tk)
{
	int i;
	for (i = 0;  serv_vect[i].funct != NULL;  i++)
		if (serv_vect[i].code == tk->code)
			return serv_vect[i].funct(tid, tk);
	return NULL;
}



/* -----------------------------------------------------------------
 * Function:    PRIVATE void *service_thread (void *arg)
 *
 * Description: Main loop of a service thread
 *
 * Returns:     NULL
 */
PRIVATE void *f0m__service_thread (void *arg)
{
	request_t *req;
	tkout_t   *tko;
	int tid = (int)arg;
	for (;;) {
		req = f0m_dequeue_request (tid);
		assert (req != NULL);
		assert (req->tid == tid);
		tko = service (req->tid, req->tk); /* Execute service */
		assert (tko);
/* 		ss = tkin_2_str (req->tk); */
/* 		assert (ss); */
/* 		printf ("tid = %d\n", tid); */
/* 		printf ("size = %d\n", strlen (ss)); */
/* 		printf ("%s", ss); */
/* 		printf ("-------------------------------- TKIN -------------------------------- */
/* %s---------------------------------------------------------------------- */
/* ", ss); */
/* 		free (ss); */
		f0m_enqueue_reply (tid, req->id, tko);
	}
	assert (0);					/* This point should never be reached */
}




PRIVATE int f0m__cmp_tid_f (const void *data, const void *what, void *arg)
{
	const thread_info_t *info = data;
	int tid = (int)what;
	return info->tid == tid;
}



/* -----------------------------------------------------------------
 * Function:    start_service_thread()
 *
 * Description: Start a new the system thread
 *
 * Returns:     The status
 */
PRIVATE int f0m__start_service_thread (int tid)
{
	int err;
	thread_info_t *info;

	/* Be sure the service */
	assert (serv_q != NULL);
	assert (tid >= -1);
	
	/* Allocate data repository */
	info = malloc (sizeof (*info));
	assert (info != NULL);
	info->tid = tid;

	/* Look for tid in the list of existing threads */
	if (chain_find (serv_q, (void *)tid, f0m__cmp_tid_f, NULL) == NULL)
		chain_enqueue (serv_q, (void *)info);
/* 		msg_pid_fatal ("Tid %d is already in serv_q\n", tid); */


	/* Create a new 'serv_thread_t' structure and fill it */
	err = pthread_create (&(info->thid), pthread_attr_default,
						  f0m__service_thread, (void *)tid);
	assert (err == 0);
	return err;
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */
 
/* -----------------------------------------------------------------
 * Function:    int status = f0m_be_initialize (void)
 *
 * Description: Start the system thread
 *
 * Returns:     The status
 */
int f0m_be_initialize (void)
{
	serv_q = chain_create ();
	assert (serv_q != NULL);
	
	/* Start system thread */
	return f0m__start_service_thread (NO_TID);
}


/* -----------------------------------------------------------------
 * Function:    int status = f0m_be_terminate (void)
 *
 * Description: Terminate the system thread
 *
 * Returns:     The status
 */
int f0m_be_terminate (void)
{
	/* Terminate threads */
	/* 	pthread_kill_other_threads_np (); */
	if (serv_q != NULL)
		chain_walk (serv_q, f0m__terminate_thread, NULL);
/* 	pthread_exit(NULL); */
	return 0;
}

