/* *****************************************************************
 *
 *  $Id: f0m_queues.c,v 1.4 2002/08/14 01:30:24 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/12 22:53:31 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation for the enqueue requests and dequeue
 *				 replies services
 *
 *		f0m_queues_init ()		-- Initialization
 *		f0m_queues_terminate ()	-- Termination
 *		f0m_enqueue_request ()	-- Enqueue a request
 *		f0m_dequeue_request ()	-- Dequeue a request
 *		f0m_enqueue_reply ()		-- Enqueue a reply
 *		f0m_dequeue_reply ()		-- Dequeue a reply
 *
 * -----------------------------------------------------------------
 */

#include "config.h"
#include "glob.h"

#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include "chain.h"
#include "f0s_dbase.h"
#include "f0m_queues.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */


PRIVATE chain_t *req_q = NULL;
PRIVATE chain_t *rep_q = NULL;


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    unsigned long reqid = generate_new_request_id (void)
 *
 * Description: Generate a new request ID
 *
 * Returns:     The new request ID
 */
PRIVATE unsigned long generate_new_request_id (void)
{
	PRIVATE pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
	PRIVATE unsigned long the_id = 0;
	unsigned long reqid ;
	
	pthread_mutex_lock (&mut);
	reqid = ++the_id;
	pthread_mutex_unlock (&mut);
	return reqid;
}


/* -----------------------------------------------------------------
 * Function:    int f0m__cmp_id_f (void *data, void *args)
 *
 * Description: Test 'if (((reply_t *)data)->id == (int) args)'
 *
 * Returns:     1 if condition evaluates to TRUE
 *				0 otherwise
 */
PRIVATE int f0m__cmp_id_f (void *data, void *args)
{
	assert (data != NULL);
#ifdef EBUG
	msg_pid_write ("%s()::data->tid=%d\n", __FUNCTION__, ((reply_t *)data)->tid);
#endif
	if (((reply_t *)data)->id == (int) args)
		return 1;				/* TRUE */
	return 0;					/* FALSE */
}


/* -----------------------------------------------------------------
 * Function:    int f0m__cmp_tid_f (void *data, void *args)
 *
 * Description: Test 'if (((reply_t *)data)->tid == (int) args)'
 *
 * Returns:     1 if condition evaluates to TRUE
 *				0 otherwise
 */
PRIVATE int f0m__cmp_tid_f (void *data, void *args)
{
	assert (data != NULL);
#ifdef EBUG
	msg_pid_write ("%s()::data->tid=%d\n", __FUNCTION__, ((request_t *)data)->tid);
#endif
	if (((request_t *)data)->tid == (int) args)
		return 1;				/* TRUE */
	return 0;					/* FALSE */
}


/* -----------------------------------------------------------------
 * Function:    int f0m__true_f (void *data, void *args)
 *
 * Description: Always stop iteration
 *
 * Returns:     1 (always)
 */
PRIVATE int f0m__true_f (void *data, void *args)
{
	return 1;				/* TRUE */
}


/* -----------------------------------------------------------------
 * Function:    request_t *req = tkin_2_request (int tid,
 *												 const tkin_t *tk)
 *
 * Description: Generate a request ID and pack it with the 'tkin_t'
 *				token into a 'request_t' structure.
 *
 * Returns:     A pointer to a newly allocated 'request_t' structure
 */
PRIVATE request_t *tkin_2_request (int tid, const tkin_t *tk)
{
	request_t *req = malloc (sizeof (*req));
	assert (req != NULL);
	req->id = generate_new_request_id ();
	req->tid = tid;
	req->tk = tk;
	return req;
}


/* -----------------------------------------------------------------
 * Function:    reply_t *rep = tkout_2_reply (int tid, int id, 
 *											  const tkout_t *tk)
 *
 * Description: Pack the request ID, task ID and the reply token into
 *				a 'reply_t' structure.
 *
 * Returns:     A pointer to a newly allocated 'reply_t' structure
 */
PRIVATE reply_t *tkout_2_reply (int tid, int id, const tkout_t *tk)
{
	reply_t *rep = malloc (sizeof (*rep));
	assert (rep != NULL);
	rep->id = id;
	rep->tid = tid;
	rep->tk = tk;
	return rep;
}




/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    int f0m_queues_init (void)
 *
 * Description: Initialize queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
PUBLIC int f0m_queues_init (void)
{
	assert (req_q == NULL);
	assert (rep_q == NULL);
	req_q = chain_create ();
	rep_q = chain_create ();
	assert (req_q != NULL);
	assert (rep_q != NULL);
	return 0;
}

/* -----------------------------------------------------------------
 * Function:    int f0m_queues_terminate (void)
 *
 * Description: Terminate queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
PUBLIC int f0m_queues_terminate (void)
{
	if (req_q != NULL) chain_destroy (req_q);
	if (rep_q != NULL) chain_destroy (rep_q);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    int f0m_enqueue_request (int tid, const tkin_t *req)
 *
 * Description: 'tid'	The (symbolic) id of the target process
 *				'req'	Token with the request code and arguments
 *
 * Returns:     The request ID
 */
PUBLIC int f0m_enqueue_request (int tid, const tkin_t *tk)
{
	request_t *req;
	void *err;

	/* Check if 'tid' is valid */
	if (tid != -1  &&  f0s_dbase_get_tp_pid (tid) < 0)
		return -1;

	/* Enqueue request */
	req = tkin_2_request (tid, tk);
	assert (req_q != NULL);
#ifdef EBUG
	msg_pid_write ("Queuing request #%d\n", req->id);
#endif
	err = chain_enqueue (req_q, req);
	assert (err != NULL);
	return req->id;
}


/* -----------------------------------------------------------------
 * Function:    request_t *req f0m_dequeue_request (int tid)
 *
 * Description: 'tid'	The (symbolic) id of the target process
 *
 * Returns:     A structure 'request_t'
 */
PUBLIC request_t *f0m_dequeue_request (int tid)
{
	request_t *req;

	assert (req_q != NULL);
#ifdef EBUG
	msg_pid_write ("%s::Waiting for request\n", __FUNCTION__);
#endif
	req = (request_t *) chain_dequeue_cond (req_q, f0m__cmp_tid_f, (void *)tid);
#ifdef EBUG
	msg_pid_write ("Dequeued request #%d\n", req->id);
#endif
	return req;
}


/* -----------------------------------------------------------------
 * Function:    int repid = f0m_enqueue_reply (int tid, int id,
 *											  const tkout_t *rep)
 *
 * Description: 'tid'	The (symbolic) id of the target process
 *				'rep'	Token with the reply code and arguments
 *
 * Returns:     The status
 */
PUBLIC int f0m_enqueue_reply (int tid, int id, const tkout_t *tk)
{
	reply_t *rep;
	void *err;

	rep = tkout_2_reply (tid, id, tk);
	assert (rep_q != NULL);
#ifdef EBUG
	msg_pid_write ("Queuing reply #%d\n", rep->id);
#endif
	err = chain_enqueue (rep_q, rep);
	assert (err != NULL);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *tk = f0m_dequeue_reply (int reqid)
 *
 * Description: Get the reply to the request with ID 'reqid'
 *
 * Returns:     Pointer to the reply ('tkout_t') structure
 */
PUBLIC tkout_t *f0m_dequeue_reply (int reqid)
{
	reply_t *rep;				/* The reply structure */
	chain_walker_f cond = f0m__cmp_id_f; /* The default value */

	assert (rep_q != NULL);
	assert (reqid >= -1);
/* 	if (reqid == -1) */
/* 		cond = f0m__true_f; */
#ifdef EBUG
	msg_pid_write ("%s::Waiting for reply\n", __FUNCTION__);
#endif
	rep = (reply_t *) chain_dequeue_cond (rep_q, cond, (void *)reqid);
#ifdef EBUG
	msg_pid_write ("Dequeued reply #%d\n", rep->id);
#endif
	return (tkout_t *)rep->tk;
}
