/* *****************************************************************
 *
 *  $Id: f0m_tkin.c,v 1.5 2002/09/10 15:54:32 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:14:42 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation of the services associated with the
 *				 'Input Token'.
 *
 *		f0m_tkin_new ():			Create a new input token
 *		f0m_tkin_delete ():		Delete an input token and its contents
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "glob.h"
#include "msg.h"
#include "f0m_cmd_codes.h"
#include "f0m_tkin.h"
#include "config.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#define STRDUP(s)		((s) != NULL ? strdup(s) : NULL)



/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    tkin_t *tkin = f0m_tkin_new ()
 *
 * Description: Create a new 'Input Token'
 *				'code'	The code for the service
 *				'...'	Arguments to service 'code'
 *
 * Returns:     Pointer to the new token, allocated from the heap
 */
PUBLIC tkin_t *
f0m_tkin_new (int tid, cmd_code_t code, ...)
{
	va_list ap;
	tkin_t *tk;
	
	va_start (ap, code);
	tk = (tkin_t *)calloc (1, sizeof (*tk));
	assert (tk != NULL);
	if (tk) {
		tk->tid = tid;
		tk->code = code;
		switch (code) {
		case C_TIDS:
		case C_CLIENTS: {
			const char *options = va_arg (ap, const char *);
			tk->data.r_tids_clients.options = STRDUP (options);
			break;
		}
		case C_KILL:
			break;
		case C_ADD_DEBUGGER: {
			const char *host = va_arg (ap, const char *);
			int dbg_pid = va_arg (ap, int);
			int tp_pid = va_arg (ap, int);
			const char *cmd_channel = va_arg (ap, const char *);
			const char *rep_channel = va_arg (ap, const char *);
/* 			assert (host != NULL); */
			assert (dbg_pid > 0);
			assert (tp_pid >= -1);
			assert (cmd_channel != NULL  && *cmd_channel != '\0');
			assert (rep_channel != NULL  && *rep_channel != '\0');
			tk->data.r_add_debugger.host = STRDUP (host);
			tk->data.r_add_debugger.dbg_pid = dbg_pid;
			tk->data.r_add_debugger.tp_pid = tp_pid;
			tk->data.r_add_debugger.cmd_channel = STRDUP (cmd_channel);
			tk->data.r_add_debugger.rep_channel = STRDUP (rep_channel);
			break;
		}
		case C_ATTACH: {
			int tp_pid = va_arg (ap, int);
			const char *host = va_arg (ap, const char *);
			tk->data.r_attach.tp_pid = tp_pid;
			tk->data.r_attach.host = STRDUP (host);
			break;
		}
		case C_DETACH:
			break;
		case C_SYMBOL_FILE: {
			const char *file = va_arg (ap, const char *);
/* 			const char *host = va_arg (ap, const char *); */
			tk->data.r_symbol_file.file = STRDUP (file);
/* 			tk->data.r_symbol_file.host = STRDUP (host); */
			break;
		}
		case C_FILE: {
			const char *file = va_arg (ap, const char *);
			const char *host = va_arg (ap, const char *);
			tk->data.r_file.file = STRDUP (file);
			tk->data.r_file.host = STRDUP (host);
			break;
		}
		case C_RUN:{
			const char *args = va_arg (ap, const char *);
			tk->data.r_run.args = STRDUP (args);
			break;
		}
		case C_STEP:
			break;
		case C_NEXT:
			break;
		case C_CONTINUE:
			break;
		case C_FINISH:
			break;
		case C_SIGNAL: {
			int tid = va_arg (ap, int);
			int signum = va_arg (ap, int);
			tk->data.r_signal.tid = tid;
			tk->data.r_signal.signum = signum;
			break;
		}
		case C_BREAK: {
			const char *file = va_arg (ap, const char *);
			int line = va_arg (ap, int);
			const char *function = va_arg (ap, const char *);
			tk->data.r_break.file = STRDUP (file);
			tk->data.r_break.line = line;
			tk->data.r_break.function = STRDUP (function);
			break;
		}
		case C_DELETE: {
			int bpid = va_arg (ap, int);
			tk->data.r_delete.bpid = bpid;
			break;
		}
		case C_SET_VARIABLE: {
			const char *variable = va_arg (ap, const char *);
			const char *expression = va_arg (ap, const char *);
			tk->data.r_set_variable.variable = STRDUP (variable);
			tk->data.r_set_variable.expression = STRDUP (expression);
			break;
		}
		case C_EVALUATE: {
			const char *expression = va_arg (ap, const char *);
			tk->data.r_evaluate.expression = STRDUP (expression);
			break;
		}
		case C_DISPLAY: {
			const char *expression = va_arg (ap, const char *);
			tk->data.r_display.expression = STRDUP (expression);
			break;
		}
		case C_UNDISPLAY: {
			int dispid = va_arg (ap, int);
			tk->data.r_undisplay.dispid = dispid;
			break;
		}
		case C_CALL: {
			const char *expression = va_arg (ap, const char *);
			tk->data.r_call.expression = STRDUP (expression);
			break;
		}
		case C_LIST: {
			const char *file = va_arg (ap, const char *);
			int start_line = va_arg (ap, int);
			const char *function = va_arg (ap, const char *);
			int n_lines = va_arg (ap, int);
			tk->data.r_list.file = STRDUP (file);
			tk->data.r_list.start_line = start_line;
			tk->data.r_list.function = STRDUP (function);
			tk->data.r_list.n_lines = n_lines;
			break;
		}
		case C_TTY: {
			const char *file = va_arg (ap, const char *);
			tk->data.r_tty.file = STRDUP (file);
			break;
		}
		case C_UP:
			break;
		case C_DOWN:
			break;
		case C_FRAME: {
			int frame = va_arg (ap, int);
			tk->data.r_frame.frame = frame;
			break;
		}
		case C_INFO_LOCALS: 
		case C_INFO_ARGS:
			break;
		case C_INFO_LINE: {
			const char *file = va_arg (ap, const char *);
			int line = va_arg (ap, int);
			const char *function = va_arg (ap, const char *);
			tk->data.r_info_line.file = STRDUP (file);
			tk->data.r_info_line.line = line;
			tk->data.r_info_line.function = STRDUP (function);
			break;
		}
		case C_INFO_PROGRAM:
			break;
		case C_INFO_STACK:
			break;
		case C_INFO_BREAK:
			break;
		case C_INFO_DISPLAY:
			break;
		case C_THREAD: {
			int threadID = va_arg (ap, int);
			tk->data.r_thread.threadID = threadID;
			break;
		}
		case C_INFO_THREADS:
			break;
		case C_SENDTO: {
			const char *command = va_arg (ap, const char *);
			tk->data.r_sendto.command = STRDUP (command);
			break;
		}
		case C_TERM_SERVER: {
			int halt = va_arg (ap, int);
			tk->data.r_terminate.halt = halt;
			break;
		}
		case C_REGISTER_CLIENT: {
			int cli_pid = va_arg (ap, int);
			const char *cli_filename = va_arg (ap, const char *);
			assert (cli_filename != NULL);
			tk->data.r_register_client.pid = cli_pid;
			tk->data.r_register_client.filename = STRDUP (cli_filename);
			break;
		}	
		default:
			msg_pid_fatal ("%s : %d : Invalid command code = %d\n",
						   __FILE__, __LINE__, code);
			break;
		}
	}
	va_end (ap);
	return tk;
}



/* -----------------------------------------------------------------
 * Function:    void f0m_tkin_delete ()
 *
 * Description: Delete an 'Input Token' and itds its contents
 *
 * Returns:     []
 */
void
f0m_tkin_delete (tkin_t *tk)
{
	if (tk != NULL)
		switch (tk->code) {
		case C_TIDS:
		case C_CLIENTS:
			free ((char *)tk->data.r_tids_clients.options);
			break;
		case C_KILL:
			break;
		case C_ADD_DEBUGGER:
			free ((char *)tk->data.r_add_debugger.host);
			free ((char *)tk->data.r_add_debugger.cmd_channel);
			free ((char *)tk->data.r_add_debugger.rep_channel);
			break;
		case C_ATTACH:
			break;
		case C_DETACH:
			break;
		case C_SYMBOL_FILE:
			free ((char *)tk->data.r_symbol_file.file);
			break;
		case C_FILE:
			free ((char *)tk->data.r_file.file);
			break;
		case C_RUN:
			break;
		case C_STEP:
			break;
		case C_NEXT:
			break;
		case C_CONTINUE:
			break;
		case C_FINISH:
			break;
		case C_SIGNAL:
			break;
		case C_BREAK:
			free ((char *)tk->data.r_break.file);
			free ((char *)tk->data.r_break.function);
			break;
		case C_DELETE:
			break;
		case C_SET_VARIABLE:
			free ((char *)tk->data.r_set_variable.variable);
			free ((char *)tk->data.r_set_variable.expression);
			break;
		case C_EVALUATE:
			free ((char *)tk->data.r_evaluate.expression);
			break;
		case C_DISPLAY:
			free ((char *)tk->data.r_display.expression);
			break;
		case C_UNDISPLAY:
			break;
		case C_CALL:
			free ((char *)tk->data.r_call.expression);
			break;
		case C_LIST:
			free ((char *)tk->data.r_list.file);
			free ((char *)tk->data.r_list.function);
			break;
		case C_TTY:
			free ((char *)tk->data.r_tty.file);
			break;
		case C_UP:
			break;
		case C_DOWN:
			break;
		case C_FRAME:
			break;
		case C_INFO_LOCALS: 
		case C_INFO_ARGS:
			break;
		case C_INFO_LINE:
			free ((char *)tk->data.r_info_line.file);
			free ((char *)tk->data.r_info_line.function);
			break;
		case C_INFO_PROGRAM:
			break;
		case C_INFO_STACK:
			break;
		case C_INFO_BREAK:
			break;
		case C_INFO_DISPLAY:
			break;
		case C_THREAD:
		case C_INFO_THREADS:
			break;
		case C_SENDTO:
			free ((char *)tk->data.r_sendto.command);
			break;
		case C_TERM_SERVER:
			break;
		case C_REGISTER_CLIENT: {
			free ((char *)tk->data.r_register_client.filename);
			break;
		}
		default:
			msg_pid_fatal ("%s : %d : Invalid command code = %d\n",
						   __FILE__, __LINE__, tk->code);
			break;
		}
	free (tk);
}
