/* *****************************************************************
 *
 *  $Id: f0s_getdelim.c,v 1.1.1.1 2002/04/28 22:06:51 jml Exp $
 *
 *  Copywright (c) 1999 by Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *  Time-stamp: <99/06/28 16:58:59 jml>
 *
 *  Universidade Nova de Lisboa
 *  Faculdade de Ciencias e Tecnologia
 *  Departamento de Informatica
 *  Parallel and Distributed Processing Group
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "f0s_getdelim.h"


#define BLOCK_SIZE 512

typedef struct buf_s buf_t;
struct buf_s {
	char  *mem;
	int    size;
	int    used;
	buf_t *next;
};


/* Create a new buffer */
buf_t *buf_new (void)
{
	buf_t *b = malloc (sizeof (*b));
	assert (b != NULL);
	b->mem = malloc (BLOCK_SIZE * sizeof (char));
	b->size = BLOCK_SIZE;
	b->used = 0;
	b->next = NULL;
	return b;
}

/* Delete a buffer */
void buf_delete (buf_t *b)
{
	buf_t *buf;
	
	assert (b != NULL);
	buf = b;
	do {
		free (buf->mem);
		b = buf->next;
		free (buf);
		buf = b;
	} while (buf != NULL);
}


/* Put a character in the buffer */
static char buf_putc (buf_t *b, char c)
{
	assert (b != NULL);
	while (b->next != NULL)
		b = b->next;
	if (b->used == b->size) {
		b->next = buf_new ();
		b = b->next;
	}
	b->mem[b->used] = c;
	b->used++;
	return c;
}

/* Get a malloc'ed copy the buffer contents (add a terminating 0 to
 * prevent the case when the contents is a string */
static char *buf_mem (buf_t *b)
{
	char *mem;
	char *memp;
	buf_t *buf;
	int tot_size;
	
	assert (b != NULL);
	tot_size = 0;
	for (buf = b;  buf != NULL;  buf = buf->next)
		tot_size += buf->used;
	mem = malloc ((tot_size + 1) * sizeof (char));
	assert (mem != NULL);
	memp = mem;
	for (buf = b;  buf != NULL;  buf = buf->next) {
		memcpy (memp, buf->mem, buf->used);
		memp += buf->used;
	}
	*memp = '\0';
	return mem;
}

/* Test for a string in the end of buffer -- this implementation is
 * terribly inefficient */
int buf_last_chars (buf_t *b, const char *delim, int delimsize)
{
	char *m;
	buf_t *p = NULL;
	
	assert (b != NULL);
	assert (delimsize <= BLOCK_SIZE);
	m = alloca ((delimsize + 1) * sizeof (char));
	assert (m != NULL);
	while (b->next != NULL) {
		p = b;
		b = b->next;
	}
	if (b->used >= delimsize) 
		/* There are enough characters in this block, copy the last
		 * 'delimsize' characters */
		memcpy (m, b->mem + b->used - delimsize, delimsize);
	else if (p != NULL) {
		/* Insufficient number of characters in this block, join the
		 * necessary characters from previous block */
		memcpy (m, p->mem + p->used + b->used - delimsize,
				delimsize - b->used);
		memcpy (m + delimsize - b->used, b->mem, b->used);
	}
	else
		/* Not enough characters in buffer */
		return 0;
	m[delimsize] = '\0';
	return (strcmp (m, delim) == 0);
}




/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    char *get_delim (IN FILE *fd, IN const char *delim)
 *
 * Description: Read chars from 'fd' until 'delim' is read.
 *				'delim' is not included in the string read.
 *
 * Returns:     Pointer to dynamically allocated string.
 *				The user should 'free()' the string when no longer needed.
 */
char *get_delim (FILE *fd, const char *delim)
{
	buf_t *buffer;
	char *str;
    int offset;
    int c;
	int delimsize;

	assert (fd != NULL);
	assert (delim != NULL);
	
	delimsize = strlen (delim);

	buffer = buf_new ();
	
    /* initialise buffer */
    offset = 0;

    for (;;) {
		c = getc (fd);
		
		if (c == 0)		/* silently drop NULs */
			continue;

		if (c == EOF) {
			if (offset == 0)
				/* If this is the first char read on this call,
				   then return EOF */
				return NULL;
			/* Else unexpected EOF: NUL-terminate buffer and return */
			break;
		}
		
		/* A normal character: add it to the buffering */
		buf_putc (buffer, c);
		offset++;

		/* Test for 'delim' in the end of the buffering */
		if (buf_last_chars (buffer, delim, delimsize))
			break;
    }
	str = buf_mem (buffer);
	str[offset - delimsize] = '\0';
	return str;
}


/* For testing purposes only... */
#if TEST_MAIN
main ()
{
	char *buffer;
	char delim[80];
	
	printf ("Delim: ");
	scanf ("%s", delim);

	printf ("Reading from stdin until \"%s\" is found\n", delim);
	buffer = get_delim (stdin, delim);
	if (buffer == NULL) buffer = "NULL was returned!!!";
	puts ("\n------------------------------------------------------------------------------");
	puts (buffer);
	puts ("\n------------------------------------------------------------------------------");
}
#endif
