/* *****************************************************************
 *
 *  $Id: f0s_lib.c,v 1.4 2002/08/12 00:29:35 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/09 19:46:51 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <assert.h>

#include "chain.h"

#include "f0s_cmd_gen.h"
#include "f0s_dbg_io.h"
#include "f0s_parse.h"
#include "f0s_dbase.h"
#include "f0s_lib.h"

#include "config.h"


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

 

/* -----------------------------------------------------------------
 * Function:    is_valid_tp_pid()
 *
 * Description: Check if 'pid' is a valid process identifier in the
 *				local machine.
 *				For this, try to open the file "/proc/<pid>/exe".
 *
 * Returns:     TRUE if yes, FALSE if no.
 */
static int is_valid_tp_pid (int tp_pid)
{
#define PATH_MAX 256
	char path[PATH_MAX];

	sprintf (path, "/proc/%d/exe", tp_pid);

	/* Check if 'file' is a regular file with execution permission */
	/* 	status = stat (path, &sbuf); */
	/* 	if (status < 0) return 0; */
	/* 	if (!(S_ISREG(sbuf.st_mode))) return 0; */ /* BUG -- must check
													* permissions too */ 
	return 1;
}



/* -----------------------------------------------------------------
 * Function:    attach_tid()
 *
 * Description: Attach a component-level debugger to a process runing
 *				in the same machine.  Assumes 'tp_pid' and 'dbg_id'
 *				are valid identifiers.
 *
 * Returns:     The 'Output Token'.
 */
static tkout_t *attach_tid (int tid, int tp_pid)
{
	char *command;
	char *reply;
	tkout_t *attach;

	/* Test if 'tid' is a valid Debugger Identifier and is not being 
	 * used already */
	if ( ! f0s_dbase_is_unused (tid))
		return f0s_tkout_new (E_RTS_INVALID_TID);	/* 'tid' is non-existent or being used */

	/* Build the command string */
	command = f0s_gen_command (CMD_ATTACH, tp_pid);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);
	
	attach = f0s_parse_output (tid, CMD_ATTACH, reply);
	/* On success map the 'tid' to 'tp_pid' */
	if (attach->status == 0  &&  attach->err == NULL
		&&  attach->rep != NULL) {
		rattach_t *att = (rattach_t *)chain_first_data (attach->rep->data);
		assert (att != NULL);
		att->tid = tid;
		att->tp_pid = tp_pid;
		/* Mark 'tid' as being used with 'tp_pid' */
		f0s_dbase_set_tp_pid (tid, tp_pid);
		f0s_dbase_set_attached (tid, 1);
	}

	/* Return the Output Token (or NULL if error) */
	return attach;
}


/* -----------------------------------------------------------------
 * Function:    cmd_zero_args()
 *
 * Description: Send a command with no arguments to the CLD
 *
 * Returns:     The status of the request
 */
static tkout_t *cmd_zero_args (int tid, cmd_code_t code)
{
	char *command;
	char *reply;
	
	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (code, tid);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, code, reply);
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

 
/* -----------------------------------------------------------------
 * Function:    f0s_initialize()
 *
 * Description: Initialize all the necessary data structures
 *
 * Returns:     The status
 */
int f0s_initialize (void)
{
	return 0;	
}


/* -----------------------------------------------------------------
 * Function:    f0s_terminate()
 *
 * Description: Terminate the debugging in this node.  Detach from all 
 *				processes and kil all CLDs.
 *
 * Returns:     The status
 */
int f0s_terminate (void)
{
	f0s_dbase_cleanup ();
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    f0s_tids()
 *
 * Description: Get a list of all existing process TIDs.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_tids (const char *options)
{
	int tid;
	int all = 0;
	tkout_t *tids;

	/* Be sure 'options' is not NULL */
	if (options == NULL)
		options = "";
	/* Check for '-a' option */
	else if (strstr (options, "-a") != NULL)
		all = 1;
	else if (*options != '\0')
		return f0s_tkout_new (E_RTS_INVALID_OPTION); 

	/* Create output token and list of tids */
	tids = f0s_tkout_new (E_RTS_OK);
	if (tids == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR); /* Memory error */
	tids->tid = 0;
	tids->str = NULL;

	tids->rep = f0s_tkout_data_new ();	
	if (tids->rep == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR); /* Memory error */
	tids->rep->str = NULL;
	tids->rep->code = R_TIDS;
	tids->rep->data = f0s_dbase_list ();

	/* Return the Output Token (or NULL if error) */
	return tids;				/* OK */
}


/* -----------------------------------------------------------------
 * Function:    f0s_kill()
 *
 * Description: Kill a process being debugger but let the CLD
 *				    available for a next usage
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_kill (int tid)
{
	char *command;
	char *reply;
	tkout_t *kill;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_KILL);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);
	
	kill = f0s_parse_output (tid, CMD_KILL, reply);
	if (kill->status == 0  &&  kill->err == NULL)
		/* Mark 'tid' as being unused */
		f0s_dbase_delete (tid);

	/* Return the Output Token (or NULL if error) */
	return kill;
}


/* -----------------------------------------------------------------
 * Function:    f0s_add_debugger ()
 *
 * Description: Add a 'ready to be controlled' debugger to Fiddle
 *				'dbg_pid' should contain the PID of the debugger
 *				'tp_pid' should contain the PID of the target process
 *						 or -1 if process is not loaded or not running yet
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_add_debugger (int dbg_pid, int tp_pid,
						   const char *cmd_channel, const char *rep_channel)
{
	tkout_t *tkout;
	int tid;
	radddebugger_t *info;
	void *status;

	tid = f0s_dbase_add_cld (dbg_pid, tp_pid, cmd_channel, rep_channel);
	if (tid < 0 )
		return f0s_tkout_new (E_RTS_INVALID_ARGUMENT);
	tkout = f0s_tkout_new (E_RTS_OK);
	if (tkout == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR);
	tkout->tid = tid;
	tkout->rep = f0s_tkout_data_new();
	if (tkout->rep == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR);
	tkout->rep->code = R_ADD_DEBUGGER;

	/* Fill reply data */
	info = malloc (sizeof (*info));
	if (info == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR);
	info->tid = tid;
	status = chain_enqueue (tkout->rep->data, info);
	assert (status != NULL);

	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    f0s_attach()
 *
 * Description: Find an unsused component-level debugger or start one
 *				if none was found.  Then attach the debugger to the
 *				process 'tp_pid'.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_attach (int tp_pid)
{
	int tid;

	/* Test if 'tp_pid' is a valid process identifier, i.e. if there
	 * is a process running in the local machine with this PID */
	if (! is_valid_tp_pid (tp_pid))
		return f0s_tkout_new (E_RTS_INVALID_TP_PID); /* Invalid 'tp_pid' */

	/* Find an unused component-level debugger */
	tid = f0s_dbase_find_unused ();
	if ( tid < 0 )
		return f0s_tkout_new (E_RTS_NO_CLD_AVAILABLE); /* Problems finding/starting a CLD */

	/* Attach 'tid' to 'tp_pid' */
	return attach_tid (tid, tp_pid);
}


/* -----------------------------------------------------------------
 * Function:    f0s_detach()
 *
 * Description: Detach a component-level debugger from the process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_detach (int tid)
{
	char *command;
	char *reply;
	tkout_t *detach;
	int tp_pid;

	/* Check if 'tid' is valid */
	tp_pid = f0s_dbase_get_tp_pid (tid);
	if (tp_pid < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_DETACH);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	detach = f0s_parse_output (tid, CMD_DETACH, reply);
	if (detach->status == 0  &&  detach->err == NULL 
		&& detach->rep->data != NULL) {
		rdetach_t *d = chain_first_data (detach->rep->data);
		assert (d != NULL);
		d->tp_pid = tp_pid;
		/* Mark 'tid' as being unused */
		f0s_dbase_delete (tid);
	}

	/* Return the Output Token (or NULL if error) */
	return detach;
}


/* -----------------------------------------------------------------
 * Function:    f0s_symbol_file()
 *
 * Description: Load the symbols
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_symbol_file (int tid, const char *file)
{
	char *command;
	char *reply;
	struct stat sbuf;
	int status;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Check if 'file' is a regular file with execution permission */
	status = stat (file, &sbuf);
	if (status < 0)
		return f0s_tkout_new (E_RTS_STAT_FILE);
	if (!(S_ISREG(sbuf.st_mode)))
		return f0s_tkout_new (E_RTS_FILE_PERM);	/* BUG -- must check permissions too */ 

	/* Build the command string */
	command = f0s_gen_command (CMD_SYMBOL_FILE, file);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_SYMBOL_FILE, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_file()
 *
 * Description: Load a process image into memory
 *				'tid == -1' load a new process image
 *				'tid != -1' reload a process image
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_file (int tid, const char *file)
{
	char *command;
	char *reply;
	struct stat sbuf;
	int status;
	tkout_t *tk;

	/* Check if 'file' is a regular file with execution permission */
	status = stat (file, &sbuf);
	if (status < 0)
		return f0s_tkout_new (E_RTS_STAT_FILE);
	if (! ((S_ISREG(sbuf.st_mode))  &&  (sbuf.st_mode & S_IXUSR)))
		return f0s_tkout_new (E_RTS_FILE_PERM);	/* BUG -- must check permissions too */ 

	/* If not reusing a TID, find an unused component-level debugger */
	if (tid == -1)
		tid = f0s_dbase_find_unused ();
	if ( tid < 0 )
		return f0s_tkout_new (E_RTS_NO_CLD_AVAILABLE); /* Problems finding/starting a CLD */

	/* Build the command string */
	command = f0s_gen_command (CMD_FILE, file);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	tk = f0s_parse_output (tid, CMD_FILE, reply);
	/* On success set the 'tid' */
	if (tk->status == 0  &&  tk->rep != NULL) {
		/* Set the 'tid' in the 'TkOut' structure */
		rfile_t *file = (rfile_t *)chain_first_data (tk->rep->data);
		assert (file != NULL);
		file->tid = tid;
		/* Set the target process PID to the value of 0 */
		f0s_dbase_set_tp_pid (tid, 0);
		f0s_dbase_set_attached (tid, 0);
	}

	/* Return the Output Token (or NULL if error) */
	return tk;
}


/* -----------------------------------------------------------------
 * Function:    f0s_run()
 *
 * Description: Start the execution of a process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_run (int tid, const char *args)
{
	char *command;
	char *reply;
	tkout_t *run;
	

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_RUN, args);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Return immediately if there were problems getting the reply */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Build the Output Token (or NULL if error) */
	run = f0s_parse_output (tid, CMD_RUN, reply);

	if (run->status >= 0  &&  run->err == NULL) {
		/* There were no errors, set the TP_PID in the database */
		tkout_t *info = f0s_info_program (tid);
		if (info->status == 0  &&  info->err == NULL
			&&  info->rep != NULL) {
			rinfoprogram_t *ip = (rinfoprogram_t *)chain_first_data (info->rep->data);
			assert (ip != NULL);
			f0s_dbase_set_tp_pid (tid, ip->tp_pid);
			f0s_tkout_delete (info);
		}
	}
	return run;
}

/* -----------------------------------------------------------------
 * Function:    f0s_step()
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps into the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_step (int tid)
{
	return cmd_zero_args (tid, CMD_STEP);
}


/* -----------------------------------------------------------------
 * Function:    f0s_next()
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps over the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_next (int tid)
{
	return cmd_zero_args (tid, CMD_NEXT);
}


/* -----------------------------------------------------------------
 * Function:    f0s_continue()
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_continue (int tid)
{
	return cmd_zero_args (tid, CMD_CONTINUE);
}


/* -----------------------------------------------------------------
 * Function:    f0s_signal()
 *
 * Description: Send a signal to a process.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_signal (int tid, int signum)
{
	int tp_pid;
	

	/* Check if 'tid' is valid */
	tp_pid = f0s_dbase_get_tp_pid (tid);
	if (tp_pid < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Ensure <tid> is already running */
	if (tp_pid == 0)
		return f0s_tkout_new (E_NOT_RUN);

	/* Check if 'signum' is valid */
	if (signum < 1  ||  signum >= _NSIG)
		return f0s_tkout_new (E_RTS_INVALID_SIGNUM);

	/* Send the signal to the process */
	if (kill (tp_pid, signum) < 0)
		return f0s_tkout_new (E_RTS_SIGNAL);

	/* Return OK */
	return f0s_tkout_new (E_RTS_OK);
}


/* -----------------------------------------------------------------
 * Function:    f0s_finish()
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_finish (int tid)
{
	return cmd_zero_args (tid, CMD_FINISH);
}


/* -----------------------------------------------------------------
 * Function:    f0s_break()
 *
 * Description: Set a breakpoint.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_break (int tid, const char *file, int line, const char *function)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Check if one and only one of 'line' and 'function' are present */
	if ( (line != -1  &&  function != NULL)
		 || (line == -1  &&  function == NULL))
		return f0s_tkout_new (E_RTS_ONE_LINE_FUNCTION); /* None or both defined */

	/* Build the command string */
	command = f0s_gen_command (CMD_BREAK, file, line, function);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_BREAK, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_delete()
 *
 * Description: Remove a breakpoint.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_delete (int tid, int bpid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_DELETE, bpid);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_DELETE, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_set_variable()
 *
 * Description: Set 'variable' to 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_set_variable (int tid, const char *variable,
						   const char *expression)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	if (variable == NULL  ||  strlen (variable) == 0
		|| expression == NULL  ||  strlen (expression) == 0)
		return f0s_tkout_new (E_RTS_INVALID_ARGUMENT);

	/* Build the command string */
	command = f0s_gen_command (CMD_SET_VARIABLE, variable, expression);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_SET_VARIABLE, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_evaluate()
 *
 * Description: Evaluate an 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_evaluate (int tid, const char *expression)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	if (expression == NULL  ||  strlen (expression) == 0)
		return f0s_tkout_new (E_RTS_INVALID_ARGUMENT);
	command = f0s_gen_command (CMD_EVALUATE, expression);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_EVALUATE, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_locals()
 *
 * Description: Get name and value of local variables 
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_locals (int tid)
{
	return cmd_zero_args (tid, CMD_INFO_LOCALS);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_args()
 *
 * Description: Get name and value of function arguments
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_args (int tid)
{
	return cmd_zero_args (tid, CMD_INFO_ARGS);
}


/* -----------------------------------------------------------------
 * Function:    f0s_display()
 *
 * Description: Display 'expression' every time the program stops
 *				If 'expression == NULL' list list all expressions/values
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_display (int tid, const char *expression)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	if (expression == NULL)
		expression = "";
	command = f0s_gen_command (CMD_DISPLAY, expression);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_DISPLAY, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_undisplay()
 *
 * Description: Remove a display expression
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_undisplay (int tid, int dispid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_UNDISPLAY, dispid);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_UNDISPLAY, reply);
}


/* -----------------------------------------------------------------
 * Function:	tkof0s_call()
 *
 * Description: Call a function and get it's return value
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_call (int tid, const char *expression)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	if (expression == NULL  ||  strlen (expression) == 0)
		return f0s_tkout_new (E_RTS_INVALID_ARGUMENT);
	command = f0s_gen_command (CMD_CALL, expression);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_CALL, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_list()
 *
 * Description: List a source file.  
 *              The arguments 'start_line' and 'function' are exclusive.
 *              The arguments 'n_lines' is optional.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_list (int tid, const char *file, int start_line,
				   const char *function, int n_lines)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Check if one and only one of 'line' and 'function' are present */
	if ( (start_line != -1  &&  function != NULL)
		 || (start_line == -1  &&  function == NULL))
		return f0s_tkout_new (E_RTS_ONE_LINE_FUNCTION); /* None or both defined */

	/* If 'function' was given, determine the starting line */
	if (function != 0) {
		rinfoline_t *il;
		tkout_t *info = f0s_info_line (tid, file, -1, function);
		if (info->status != FIDDLE_OK  ||  info->err != NULL
			||  info->rep == NULL)
			return info; /* Some error occurred */
		il = (rinfoline_t *)chain_first_data (info->rep->data);
		start_line = il->line - 1;
		f0s_tkout_delete (info);
		if (start_line <= 0)
			return f0s_tkout_new (E_RTS_INVALID_LINE_NUMBER); /* Error! */
	}

	/* Build the command string */
	command = f0s_gen_command (CMD_LIST, file, start_line, NULL, n_lines);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_LIST, reply);
}


/* -----------------------------------------------------------------
 * Function:	tkof0s_tty()
 *
 * Description: Set the I/O tty for the process being debugged
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_tty (int tid, const char *file)
{
	char *command;
	char *reply;
	struct stat sbuf;
	int status;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Check if 'file' is a regular file with execution permission */
	status = stat (file, &sbuf);
	if (status < 0)
		return f0s_tkout_new (E_RTS_STAT_TTY);
	if (!(S_ISCHR(sbuf.st_mode)))
		return f0s_tkout_new (E_RTS_TTY_PERM);	/* BUG -- must check permissions too */ 

	/* Build the command string */
	command = f0s_gen_command (CMD_TTY, file);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_TTY, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_line()
 *
 * Description: Get information about a source line.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_line (int tid, const char *file, int line,
						const char *function)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Check if one and only one of 'line' and 'function' are present */
	if ( (line != -1  &&  function != NULL))
		return f0s_tkout_new (E_RTS_ONE_LINE_FUNCTION); /* None or both defined */

	/* Build the command string */
	command = f0s_gen_command (CMD_INFO_LINE, file, line, function);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_INFO_LINE, reply);
}


/* -----------------------------------------------------------------
 * Function:	tkof0s_info_program()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_program (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_INFO_PROGRAM);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_INFO_PROGRAM, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_stack()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_stack (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_INFO_STACK);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_INFO_STACK, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_up()
 *
 * Description: Go "up" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_up (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_UP);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_UP, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_down()
 *
 * Description: Go "down" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_down (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_DOWN);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_DOWN, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_frame()
 *
 * Description: Select a specific stack frame
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_frame (int tid, int frame)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_FRAME, frame);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_FRAME, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_break()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_break (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_INFO_BREAK);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_INFO_BREAK, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_display()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_display (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_INFO_DISPLAY);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_INFO_DISPLAY, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_thread()
 *
 * Description: Switch the current thread.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_thread (int tid, int threadID)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_THREAD, threadID);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_THREAD, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_info_threads()
 *
 * Description: Get information about the existing threads.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_info_threads (int tid)
{
	char *command;
	char *reply;

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);

	/* Build the command string */
	command = f0s_gen_command (CMD_INFO_THREADS);
	if (command == NULL)
		/* Only god knows why this may fail ! */
		return f0s_tkout_new (E_RTS_COMMAND_GENERATION);

	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	free (command);

	/* Build the Output Token if command was successefully sent */
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the Output Token (or NULL if error) */
	return f0s_parse_output (tid, CMD_INFO_THREADS, reply);
}


/* -----------------------------------------------------------------
 * Function:    f0s_sendto()
 *
 * Description: Send a command string directly to the debugger.
 *				'tkout' contains the output (NOT PARSED).
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0s_sendto (int tid, const char *command)
{
	char *reply;

	/* Be sure 'command' is not empty */
	if (command == NULL)
		return f0s_tkout_new (E_RTS_NO_COMMAND);

	/* Check if 'tid' is valid */
	if (f0s_dbase_get_tp_pid (tid) < 0)
		return f0s_tkout_new (E_RTS_INVALID_TID);
	
	/* Send command to debugger and get its reply */
	reply = f0s_send_receive (tid, command);
	if (reply == NULL)
		return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Build the Output Token */
	return f0s_parse_output (tid, CMD_SENDTO, reply);
}

