/* *****************************************************************
 *
 *  $Id: f1m_back_end.c,v 1.10 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/27 00:00:16 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation of the back-end of the services
 *				 provided by FIDDLE-1.
 *
 *		f1m_be_initialize ()	-- Initialize FIDDLE-1
 *		f1m_be_terminate ()		-- Terminate FIDDLE-1
 *		f1m_be_is_valid_tid ()	-- Check a given TID is valid
 *
 * -----------------------------------------------------------------
 */

#include "config.h"
#include "glob.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>

#include <netdb.h>

#include "chain.h"

#include "f0m_tkin.h"
#include "f0m_rw_to_file.h"

#include "f0m_lib.h"

#include "f1m_queues.h"
#include "f1m_servdata.h"
#include "f1m_back_end.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */


typedef tkout_t *(*serv_f)(const tkin_t *tk);


typedef struct serv_vect_s serv_vect_t;
struct serv_vect_s {
	int     code;
	serv_f	funct;
};


PRIVATE tkout_t *serv_tids (const tkin_t *tk);
PRIVATE tkout_t *serv_kill (const tkin_t *tk);
PRIVATE tkout_t *serv_add_debugger (const tkin_t *tk);
PRIVATE tkout_t *serv_attach (const tkin_t *tk);
PRIVATE tkout_t *serv_detach (const tkin_t *tk);
PRIVATE tkout_t *serv_symbol_file (const tkin_t *tk);
PRIVATE tkout_t *serv_file (const tkin_t *tk);
PRIVATE tkout_t *serv_run (const tkin_t *tk);
PRIVATE tkout_t *serv_step (const tkin_t *tk);
PRIVATE tkout_t *serv_next (const tkin_t *tk);
PRIVATE tkout_t *serv_continue (const tkin_t *tk);
PRIVATE tkout_t *serv_finish (const tkin_t *tk);
PRIVATE tkout_t *serv_signal (const tkin_t *tk);
PRIVATE tkout_t *serv_break (const tkin_t *tk);
PRIVATE tkout_t *serv_delete (const tkin_t *tk);
PRIVATE tkout_t *serv_set_variable (const tkin_t *tk);
PRIVATE tkout_t *serv_evaluate (const tkin_t *tk);
PRIVATE tkout_t *serv_display (const tkin_t *tk);
PRIVATE tkout_t *serv_undisplay (const tkin_t *tk);
PRIVATE tkout_t *serv_call (const tkin_t *tk);
PRIVATE tkout_t *serv_list (const tkin_t *tk);
PRIVATE tkout_t *serv_tty (const tkin_t *tk);
PRIVATE tkout_t *serv_up (const tkin_t *tk);
PRIVATE tkout_t *serv_down (const tkin_t *tk);
PRIVATE tkout_t *serv_frame (const tkin_t *tk);
PRIVATE tkout_t *serv_info_locals (const tkin_t *tk);
PRIVATE tkout_t *serv_info_args (const tkin_t *tk);
PRIVATE tkout_t *serv_info_line (const tkin_t *tk);
PRIVATE tkout_t *serv_info_program (const tkin_t *tk);
PRIVATE tkout_t *serv_info_stack (const tkin_t *tk);
PRIVATE tkout_t *serv_info_break (const tkin_t *tk);
PRIVATE tkout_t *serv_info_display (const tkin_t *tk);
PRIVATE tkout_t *serv_thread (const tkin_t *tk);
PRIVATE tkout_t *serv_info_threads (const tkin_t *tk);
PRIVATE tkout_t *serv_tkin (const tkin_t *tk);
PRIVATE tkout_t *serv_sendto (const tkin_t *tk);




PRIVATE serv_vect_t serv_vect[] =
{
	{C_TIDS,			serv_tids},
	{C_KILL,			serv_kill},
	{C_ADD_DEBUGGER,	serv_add_debugger},
	{C_ATTACH,			serv_attach},
	{C_DETACH,			serv_detach},
	{C_SYMBOL_FILE,	serv_symbol_file},
	{C_FILE,			serv_file},
	{C_RUN,		    serv_run},
	{C_STEP,			serv_step},
	{C_NEXT,			serv_next},
	{C_CONTINUE,		serv_continue},
	{C_FINISH,			serv_finish},
	{C_SIGNAL,			serv_signal},
	{C_BREAK,			serv_break},
	{C_DELETE,			serv_delete},
	{C_SET_VARIABLE,	serv_set_variable},
	{C_EVALUATE,		serv_evaluate},
	{C_DISPLAY,		serv_display},
	{C_UNDISPLAY,		serv_undisplay},
	{C_CALL,			serv_call},
	{C_LIST,			serv_list},
	{C_TTY,		    serv_tty},
	{C_UP,				serv_up},
	{C_DOWN,			serv_down},
	{C_FRAME,			serv_frame},
	{C_INFO_LOCALS,		serv_info_locals},
	{C_INFO_ARGS,		serv_info_args},
	{C_INFO_LINE,		serv_info_line},
	{C_INFO_PROGRAM,	serv_info_program},
	{C_INFO_STACK,		serv_info_stack},
	{C_INFO_BREAK,		serv_info_break},
	{C_INFO_DISPLAY,	serv_info_display},
	{C_THREAD,			serv_thread},
	{C_INFO_THREADS,	serv_info_threads},
	{C_TKIN,			serv_tkin},
	{C_SENDTO,			serv_sendto},
	{C_LAST,			NULL},
};


PRIVATE int f1m__start_service_thread (int tid);


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


PRIVATE int
f1m__new_gtid (void)
{
	static pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
	static int the_gtid = 0;
	int gtid ;
	
	pthread_mutex_lock (&mut);
	gtid = ++the_gtid;
	pthread_mutex_unlock (&mut);
	return gtid;
}


PRIVATE int
f1m__new_servid (void)
{
	static pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
	static int the_servid = 0;
	int servid ;
	
	pthread_mutex_lock (&mut);
	servid = ++the_servid;
	pthread_mutex_unlock (&mut);
	return servid;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tcp_connect ()
 *
 * Description: Connet to a server, in 'host' at address 'serv'
 *				'host' is a name string, a dotted decimal (IPV4)  
 *					   or a hex (IPV6) address string
 *				'serv' is a service name, or a decimal (port number)
 *					   string
 *				(adapted from "Unix Networking Programming", Vol. I, 
 *				 by W. Richard Stevens)
 *
 * Returns:     The socket descriptor
 */
PRIVATE int
f1m__tcp_connect (const char *host, const char *serv)
{
	int sockfd, n;
	struct addrinfo hints, *res, *ressave;

	memset (&hints, 0, sizeof (hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	n = getaddrinfo (host, serv, &hints, &res);
	assert (n == 0);
	for (ressave = res;  res!= NULL;  res = res->ai_next) {
		sockfd = socket (res->ai_family,
						 res->ai_socktype, res->ai_protocol);
		if (sockfd > 0) {
			if (connect (sockfd, res->ai_addr, res->ai_addrlen) == 0)
				break;			/* Success!!! */
			close (sockfd);
		}
		sockfd = -1;
	}
	freeaddrinfo (ressave);
	return sockfd;
}



/* -----------------------------------------------------------------
 * Function:    f1m__getenv_or_default()
 *
 * Description: Get environment variable value, if undef use default 
 *
 * Returns:     []
 */
PRIVATE const char *
f1m__getenv_or_default (const char *env_var, const char *default_value)
{
	const char *val = getenv (env_var);
	if (val == NULL)			/* Environment variable not defined */
		val = default_value;	/* Use default value */
	return val;
}



/* -----------------------------------------------------------------
 * Function:    f1m__launch_server()
 *
 * Description: Check wich remote shell command to use (e.g., rsh,
 *				ssh), fork() tolaunch the 'f1m_server'
 *
 * Returns:     []
 */
PRIVATE int
f1m__launch_server (const char *f0mserv_mach, const char *f0mserv_service, 
					const char *argv[], char server[])
{
	int i;
	int f0mserv_fd;
	const char *rsh, *server_dir, *xdr;

	/* Coudn't connect */
#ifdef EBUG
	msg_pid_write ("Couldn't connect to FIDDLE-0m server in %s:%s\n",
				   f0mserv_mach, f0mserv_service);
#endif
	
	/* Check if user defined his own RSH command using an evironment
	   variable */
	rsh = f1m__getenv_or_default ("F1M_RSH", RSH_COMMAND);

	xdr = f1m__getenv_or_default ("FIDDLE_XDR", "jml");
	assert (xdr != NULL);
	
#ifdef EBUG
	msg_pid_write ("Starting a new FIDDLE-0m server in %s:%s\n\n",
				   f0mserv_mach, f0mserv_service);
#endif
	server_dir = f1m__getenv_or_default ("F0M_SERVER_DIR", SERVER_DIR);
	sprintf (server, "%s/%s", server_dir, F0M_SERVER);
	argv[0] = rsh;
	argv[1] = f0mserv_mach;
	argv[2] = server;
	argv[3] = "-x";
	argv[4] = xdr;
	argv[5] = NULL;
#ifdef EBUG
	printf ("Executing:");
	for (i = 0;  argv[i] != NULL; i++)
		printf (" %s", argv[i]);
	puts("");
#endif

	/* Fork lo launch the server */
	if (fork () == 0) {
		close (0);
		close (1);
		execvp ((char *)argv[0], (char **)argv);
		assert (0);
	}
	
	/* Wait for f1mserver to start for, at most, 1e5*50 usec = 5 sec */
#define N_TRIES		50
#define WAIT_TIME	1e5
	for (i = 0;  i < N_TRIES;  i++) {
		f0mserv_fd = f1m__tcp_connect (f0mserv_mach, f0mserv_service);
		if (f0mserv_fd >= 0) break;	/* Success */
		usleep (WAIT_TIME);
	}
	return(f0mserv_fd);
}



/* -----------------------------------------------------------------
 * Function:    f1m__register_client()
 *
 * Description: Register server in the database
 *
 * Returns:     The structure that was inserted into the database
 */
PRIVATE serv_data_machine_t *
f1m__register_client (int server_fd, const char *machine)
{
	int servid;
	serv_data_machine_t *data;

	servid = f1m__new_servid ();
	data = f1m_tp_add_servdata_machine (servid, machine, server_fd, server_fd);
	if (data != NULL)
		f1m__start_service_thread (servid);
	return data;
}


/* -----------------------------------------------------------------
 * Function:    f1m__find_or_start_f0mserver()
 *
 * Description: Connect to local (node) daemon, and initialize the
 *				communications (receive/send) channel
 *
 * Returns:     The I/O descriptor
 */
PRIVATE serv_data_machine_t *
f1m__find_or_start_f0mserver (const char *f0mserv_mach)
{
	const char *argv[10];
	char server[256];
	int i;
	int status;
	int f0mserv_fd;					/* The I/O channel */
	const char *rsh;
	const char *server_dir;
	const char *f0mserv_service;
	serv_data_machine_t *serv_data = NULL;
	char hostname[128];

	/* Use the official host name, not aliases */
	struct hostent *he;

	/* If no hostname or "localhost" was given as the localization of
	   the f0mserver, get real hostname */
	if (strstr (f0mserv_mach, "localhost") == f0mserv_mach) {
		status = gethostname (hostname, sizeof (hostname));
		assert (status == 0);
		f0mserv_mach = hostname;
	}

	he = gethostbyname (f0mserv_mach);
	assert (he != NULL);
	assert (he->h_name != NULL);
	f0mserv_mach = he->h_name;

	/* Look for entry in database */
	serv_data = f1m_tp_machine_2_servdata (he->h_name);

	/* Host not found.  Must start a new f0mserver in this host */
	if (serv_data == NULL) {
	
		/* If omitted, fill variables with default values */
		if (f0mserv_mach == NULL)
			f0mserv_mach = f1m__getenv_or_default ("F0M_SERVER_HOST",
												   F0M_SERVER_HOST);
		f0mserv_service = f1m__getenv_or_default ("F0M_SERVER_SERVICE_NUMBER",
												  F0M_SERVER_SERVICE_NUMBER);

		/* Connet to the server */
		f0mserv_fd = f1m__tcp_connect (f0mserv_mach, f0mserv_service);
		assert (f0mserv_fd < 0); /* Otherwise the server is already
									running, ans should have been
									found above */

		f0mserv_fd = f1m__launch_server (f0mserv_mach, f0mserv_service,
										 argv, server);

		if (f0mserv_fd >= 0) 	/* Success.  Register client. */
			serv_data = f1m__register_client (f0mserv_fd, f0mserv_mach);
		else
			msg_pid_fatal ("Couldn't launch f0mserver in %s/%s\n\tThis means serious problems!!!  :-(\n\n\n", f0mserv_mach, f0mserv_service);

	}
	assert (serv_data != NULL);
	return serv_data;
}


PRIVATE int
f1m__tids_fill_in_ltid_machine (void *data, void *args)
{
	serv_data_tid_t *st;
	rtids_t *d = (rtids_t *)data;
	serv_data_machine_t *sd = (serv_data_machine_t *)args;
	d->ltid = d->gtid;
	d->machine = strdup (sd->machine);
	st = f1m_tp_sltid_2_tiddata (sd->servid,d->ltid);
	assert (st != NULL);
	d->gtid = st->gtid;
	return 0;
}

PRIVATE int
f1m__flat_tkout_rep_data_tids (void *data, void *args)
{
	chain_t *dest;
	chain_collect_t *src_node;
	tkout_t *src_data;
	serv_data_machine_t *src_orig;
	chain_t *nn;
	
	dest = (chain_t *)args;
	assert (dest != NULL);
	src_node = (chain_collect_t *)data;
	assert (src_node != NULL);
	src_orig = chain_collect_origin (src_node);
	assert (src_orig != NULL);
	src_data = chain_collect_data (src_node);
	assert (src_data != NULL);
	assert (src_data->rep != NULL);
	assert (src_data->rep->data != NULL);
	chain_walk (src_data->rep->data, f1m__tids_fill_in_ltid_machine, src_orig);
	nn = chain_merge (dest, src_data->rep->data);
	assert (nn == dest);
	f0s_tkout_delete (src_data);
	return 0;
}


/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f1m_be_zero_args ()
 *
 * Description: Execute a service with no extra arguments
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_zero_args (int gtid, int code)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	assert (data->gtid == gtid);
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, code);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_tids ()
 *
 * Description: Get a list of all existing process TIDs.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_tids (const char *options)
{
	/* For all machines, ask for 'tids'.  Map 'ltids' to 'gtids' */
	tkin_t *tkin;
	tkout_t *tkout;
	chain_t *tmp;

	/* Create the reply token */
	tkout = f0s_tkout_new (E_RTS_OK);
	if (tkout == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR); /* Memory error */

	/* Create list of tids */
	tkout->rep = f0s_tkout_data_new ();
	if (tkout->rep == NULL)
		return f0s_tkout_new (E_RTS_MEMORY_ERROR); /* Memory error */

	/* Let's start by creating a new 'data_t' structure */
	tkout->rep->str = NULL;
	tkout->rep->code = R_TIDS;
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (NO_SERVER, C_TIDS, options);
	assert (tkin != NULL);

	/* Apply the request to all machines.
	   'tmp' will be a list of 'chain_collect_t', one of each node */
	tmp = f1m_tp_apply_to_all_machines (tkin);
	assert (tmp != NULL);
/* 	tkout->rep->data = f1m_apply_to_all_machines (tkin); */
/* 	assert (tkout->rep->data != NULL); */

	/* Convert the multiple tkout_t->rep->data into just one */
	chain_prune (tmp, f1m__flat_tkout_rep_data_tids,  tkout->rep->data);
	chain_destroy (tmp);
	
	f0m_tkin_delete (tkin);

	/* Return the Output Token */
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_kill ()
 *
 * Description: Kill a process being debugger but let the CLD
 *				    available for a next usage
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_kill (int gtid)
{
	return f1m_be_zero_args (gtid, C_KILL);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_add_debugger ()
 *
 * Description: Find an unsused component-level debugger or start one
 *				if none was found.  Then attach the debugger to the
 *				process 'tp_pid'.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_add_debugger (const char *machine, int dbg_pid, int tp_pid,
					 const char *cmd_channel, const char *rep_channel)
{
	serv_data_machine_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find the I/O channels to machine */
	data = f1m_tp_machine_2_servdata (machine);
	if (data == NULL) {
		/* No server available in that machine */
		data = f1m__find_or_start_f0mserver (machine);
		assert (data != NULL);
	}

	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (NO_SERVER, C_ADD_DEBUGGER,
						 machine, dbg_pid, tp_pid, cmd_channel, rep_channel);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Locks the mutex */
	pthread_mutex_lock(&data->notid_mutex);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, -1);

	/* unlocks the mutex, whatever the answer was */
	pthread_mutex_unlock(&data->notid_mutex);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* If the service was executed with success, add an entry to the
	 * 'tids' database */
	if (tkout->status == E_RTS_OK  &&  tkout->rep != NULL) {
		int gtid = f1m__new_gtid();
		int ltid = ((radddebugger_t *)chain_first_data (tkout->rep->data))->tid;
		serv_data_tid_t *d = f1m_tp_add_servdata_tid (gtid, data->servid, ltid,
												   data->fd[0], data->fd[1]);
		assert (d != NULL);

		/* Change the 'ltid' reference to the associated 'gtid' */
		((rattach_t *)chain_first_data (tkout->rep->data))->tid = gtid;
		tkout->tid = gtid;
	}

	/* Return the reply token */
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_attach ()
 *
 * Description: Find an unsused component-level debugger or start one
 *				if none was found.  Then attach the debugger to the
 *				process 'tp_pid'.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_attach (const char *machine, int tp_pid)
{
	serv_data_machine_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find the I/O channels to machine */
	data = f1m_tp_machine_2_servdata (machine);
	if (data == NULL) {
		/* No server available in that machine */
		data = f1m__find_or_start_f0mserver (machine);
		assert (data != NULL);
	}

	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (NO_SERVER, C_ATTACH, tp_pid, machine);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Locks the mutex */
	pthread_mutex_lock(&data->notid_mutex);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, -1);

	/* unlocks the mutex, whatever the answer was */
	pthread_mutex_unlock(&data->notid_mutex);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* If the service was executed with success, add an entry to the
	 * 'tids' database */
	if (tkout->status == E_RTS_OK  &&  tkout->rep != NULL) {
		int gtid = f1m__new_gtid();
		int ltid = ((rattach_t *)chain_first_data (tkout->rep->data))->tid;
		serv_data_tid_t *d = f1m_tp_add_servdata_tid (gtid, data->servid, ltid,
												   data->fd[0], data->fd[1]);
		assert (d != NULL);

		/* Change the 'ltid' reference to the associated 'gtid' */
		((rattach_t *)chain_first_data (tkout->rep->data))->tid = gtid;
		tkout->tid = gtid;
	}

	/* Return the reply token */
	return tkout;
}


/* -----------------------------------------------------------------
 * function:    tkout_t *rep = f1m_be_detach (int gtid)
 *
 * Description: Detach a component-level debugger from the process
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_detach (int gtid)
{
	return f1m_be_zero_args (gtid, C_DETACH);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_symbol_file ()
 *
 * Description: Load the symbols
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_symbol_file (int gtid, const char *machine, const char *file)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_SYMBOL_FILE, file, machine);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_file ()
 *
 * Description: Load a process image into memory
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_file (int gtid, const char *machine, const char *file)
{
	serv_data_machine_t *mdata;
	serv_data_tid_t *tdata;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;
	int ltid;

	if (gtid != -1) {
		/* Find data based in the 'tids' database */
		tdata = f1m_tp_gtid_2_servdata (gtid);
		if (tdata == NULL)
			return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
		mdata = f1m_tp_servid_2_servdata (tdata->servid);
		if (mdata == NULL)
			 /* Invalid gtid */
			return f0s_tkout_new (E_RTS_CANNOT_START_DAEMON);
		ltid = tdata->ltid;
	}
	else {
		/* Find data based in the 'machines' database */
		tdata = NULL;
		/* Find the I/O channels to machine */
		mdata = f1m_tp_machine_2_servdata (machine);
		if (mdata == NULL) 
			/* No server available in that machine */
			mdata = f1m__find_or_start_f0mserver (machine);
		if (mdata == NULL)
			/* Daemon launch failed */
			return f0s_tkout_new (E_RTS_CANNOT_START_DAEMON);
		ltid = NO_SERVER;
		/* lock "attach/file/add-debugger" mutex */
		pthread_mutex_lock(&mdata->notid_mutex);
	}

	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (ltid, C_FILE, file, machine);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (mdata->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (mdata->servid, ltid);
	if (ltid == NO_SERVER)
		/* unlocks the mutex */
		pthread_mutex_unlock(&mdata->notid_mutex);

	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* If the service was not directed to a specific TID and it  was
	 * executed with success, add an entry to the 'tids' database */
	if (ltid == NO_SERVER
		&&  tkout->status == E_RTS_OK  &&  tkout->rep != NULL) {
		gtid = f1m__new_gtid();
		ltid = ((rattach_t *)chain_first_data (tkout->rep->data))->tid;
		tdata = f1m_tp_add_servdata_tid (gtid, mdata->servid, ltid,
										 mdata->fd[0], mdata->fd[1]);
		assert (tdata != NULL);

		/* Change the 'ltid' reference to the associated 'gtid' */
		((rattach_t *)chain_first_data (tkout->rep->data))->tid = gtid;
		tkout->tid = gtid;
	}

	/* Return the reply token */
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_run ()
 *
 * Description: Start the execution of a process
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_run (int gtid, const char *args)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_RUN, args);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_step ()
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps into the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_step (int gtid)
{
	return f1m_be_zero_args (gtid, C_STEP);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_next ()
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps over the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_next (int gtid)
{
	return f1m_be_zero_args (gtid, C_NEXT);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_continue ()
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_continue (int gtid)
{
	return f1m_be_zero_args (gtid, C_CONTINUE);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_finish ()
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_finish (int gtid)
{
	return f1m_be_zero_args (gtid, C_FINISH);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_signal ()
 *
 * Description: Send a signal to a process.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_signal (int gtid, int signum)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_SIGNAL, signum);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_break ()
 *
 * Description: Set a breakpoint.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_break (int gtid, const char *file, int line, const char *function)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_BREAK, file, line, function);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_delete ()
 *
 * Description: Remove a breakpoint.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_delete (int gtid, int bpid)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_DELETE, bpid);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_set_variable ()
 *
 * Description: Set 'variable' to 'expression'
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_set_variable (int gtid, const char *variable, const char *expression)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_SET_VARIABLE, variable, expression);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_evaluate ()
 *
 * Description: Evaluate an 'expression'
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_evaluate (int gtid, const char *expression)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_EVALUATE, expression);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_display ()
 *
 * Description: Display 'expression' every time the program stops
 *				If 'expression == NULL' list list all expressions/values
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_display (int gtid, const char *expression)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_DISPLAY, expression);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_undisplay ()
 *
 * Description: Remove a display expression
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_undisplay (int gtid, int dispid)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_UNDISPLAY, dispid);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f1m_be_call ()
 *
 * Description: Call a function and get it's return value
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_call (int gtid, const char *expression)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_CALL, expression);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_list ()
 *
 * Description: List a source file.  
 *              The arguments 'start_line' and 'function' are exclusive.
 *              The arguments 'n_lines' is optional.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_list (int gtid, const char *file, int start_line,
			 const char *function, int n_lines)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_LIST, file, start_line,
						function, n_lines);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f1m_be_tty ()
 *
 * Description: Set the I/O tty for the process being debugged
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_tty (int gtid, const char *file)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_TTY, file);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_locals ()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_locals (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_LOCALS);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_args ()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_args (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_ARGS);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_line ()
 *
 * Description: Get information about a source line.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_line (int gtid, const char *file, int line, const char *function)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_INFO_LINE, file, line, function);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f1m_be_info_program ()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_program (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_PROGRAM);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_stack ()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_stack (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_STACK);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_up ()
 *
 * Description: Go "up" on the stack
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_up (int gtid)
{
	return f1m_be_zero_args (gtid, C_UP);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_down ()
 *
 * Description: Go "down" on the stack
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_down (int gtid)
{
	return f1m_be_zero_args (gtid, C_DOWN);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_frame ()
 *
 * Description: Select a specific stack frame
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_frame (int gtid, int frame)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_FRAME, frame);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_break ()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_break (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_BREAK);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_display ()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_display (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_DISPLAY);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_thread ()
 *
 * Description: Switch the current thread.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_thread (int gtid, int threadID)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_THREAD, threadID);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_info_threads ()
 *
 * Description: Get information about the existing threads.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_info_threads (int gtid)
{
	return f1m_be_zero_args (gtid, C_INFO_THREADS);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_sendto ()
 *
 * Description: Send a command string directly to the debugger.
 *				'tkout' contains the output (NOT PARSED).
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_sendto (int gtid, const char *command)
{
	serv_data_tid_t *data;
	tkin_t *tkin;
	tkout_t *tkout;
	int err;

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */
	
	/* Build a new 'tkin_t' token */
	tkin = f0m_tkin_new (data->ltid, C_SENDTO, command);
	assert (tkin != NULL);

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);
	f0m_tkin_delete (tkin);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f1m_be_tkin ()
 *
 * Description: Set a command already packed into a 'tkin_t' strucure
 *				to process 'gtid'.
 *
 * Returns:     The 'Output Token'.
 */
PRIVATE tkout_t *
f1m_be_tkin (int gtid, const tkin_t *tkin)
{
	serv_data_tid_t *data;
	tkout_t *tkout;
	int err;
	tkin_t *tki = (tkin_t *)tkin;

	assert (tki != NULL);

	/* Find data in the 'tids' database */
	data = f1m_tp_gtid_2_servdata(gtid);
	if (data == NULL)
		return f0s_tkout_new (E_RTS_INVALID_TID); /* Invalid gtid */

	tki->tid = data->ltid;

	/* Send it to the server */
	err = f0m_write_tkin (data->fd[1], tki);
	assert (err > 0);
	f0m_tkin_delete (tki);

	/* Get the reply */
	tkout = f1m_read_tkout (data->servid, data->ltid);
	if (tkout == NULL) return f0s_tkout_new (E_RTS_SEND_RECEIVE);

	/* Return the reply token */
	tkout->tid = data->gtid;
	return tkout;
}

/* ################################################################# */


PRIVATE tkout_t *
serv_tids (const tkin_t *tk)
{
	assert (tk->code == C_TIDS);
	return f1m_be_tids (tk->data.r_tids_clients.options);
}

PRIVATE tkout_t *
serv_kill (const tkin_t *tk)
{
	assert (tk->code == C_KILL);
	return f1m_be_kill (tk->tid);
}

PRIVATE tkout_t *
serv_add_debugger (const tkin_t *tk)
{
	assert (tk->code == C_ADD_DEBUGGER);
	return f1m_be_add_debugger (tk->data.r_add_debugger.host,
								tk->data.r_add_debugger.dbg_pid,
								tk->data.r_add_debugger.tp_pid,
								tk->data.r_add_debugger.cmd_channel,
								tk->data.r_add_debugger.rep_channel);
}

PRIVATE tkout_t *
serv_attach (const tkin_t *tk)
{
	tkout_t *rep;
	assert (tk->code == C_ATTACH);
	rep = f1m_be_attach (tk->data.r_attach.host, tk->data.r_attach.tp_pid);
	if (rep != NULL  &&  rep->status == E_RTS_OK  &&  rep->err == NULL) {
		rattach_t *att = (rattach_t *)chain_first_data (rep->rep->data);
		assert (att->tp_pid = tk->data.r_attach.tp_pid);
		assert (att->tid > 0);
		f1m__start_service_thread (att->tid);
	}
	return rep;
}

PRIVATE tkout_t *
serv_detach (const tkin_t *tk)
{
	assert (tk->code == C_DETACH);
	return f1m_be_detach (tk->tid);
}

PRIVATE tkout_t *
serv_symbol_file (const tkin_t *tk)
{
	assert (tk->code == C_SYMBOL_FILE);
	return f1m_be_symbol_file (tk->tid, tk->data.r_symbol_file.host,
							   tk->data.r_symbol_file.file);
}

PRIVATE tkout_t *
serv_file (const tkin_t *tk)
{
	tkout_t *rep;
	
	assert (tk->code == C_FILE);
	rep = f1m_be_file (tk->tid, tk->data.r_file.host, tk->data.r_file.file);
	if (rep != NULL  &&  rep->status == E_RTS_OK  &&  rep->err == NULL) {
		rfile_t *file = (rfile_t *)chain_first_data (rep->rep->data);
		assert (file->tid > 0);
		f1m__start_service_thread (file->tid);
	}
	return rep;
}

PRIVATE tkout_t *
serv_run (const tkin_t *tk)
{
	assert (tk->code == C_RUN);
	return f1m_be_run (tk->tid, tk->data.r_run.args);
}

PRIVATE tkout_t *
serv_step (const tkin_t *tk)
{
	assert (tk->code == C_STEP);
	return f1m_be_step (tk->tid);
}

PRIVATE tkout_t *
serv_next (const tkin_t *tk)
{
	assert (tk->code == C_NEXT);
	return f1m_be_next (tk->tid);
}

PRIVATE tkout_t *
serv_continue (const tkin_t *tk)
{
	assert (tk->code == C_CONTINUE);
	return f1m_be_continue (tk->tid);
}

PRIVATE tkout_t *
serv_finish (const tkin_t *tk)
{
	assert (tk->code == C_FINISH);
	return f1m_be_finish (tk->tid);
}

PRIVATE tkout_t *
serv_signal (const tkin_t *tk)
{
	assert (tk->code == C_SIGNAL);
	return f1m_be_signal (tk->data.r_signal.tid, tk->data.r_signal.signum);
}

PRIVATE tkout_t *
serv_break (const tkin_t *tk)
{
	assert (tk->code == C_BREAK);
	return f1m_be_break (tk->tid, tk->data.r_break.file,
					 tk->data.r_break.line,
					 tk->data.r_break.function);
}

PRIVATE tkout_t *
serv_delete (const tkin_t *tk)
{
	assert (tk->code == C_DELETE);
	return f1m_be_delete (tk->tid, tk->data.r_delete.bpid);
}

PRIVATE tkout_t *
serv_set_variable (const tkin_t *tk)
{
	assert (tk->code == C_SET_VARIABLE);
	return f1m_be_set_variable (tk->tid, tk->data.r_set_variable.variable,
							tk->data.r_set_variable.expression);
}

PRIVATE tkout_t *
serv_evaluate (const tkin_t *tk)
{
	assert (tk->code == C_EVALUATE);
	return f1m_be_evaluate (tk->tid, tk->data.r_evaluate.expression);
}

PRIVATE tkout_t *
serv_display (const tkin_t *tk)
{
	assert (tk->code == C_DISPLAY);
	return f1m_be_display (tk->tid, tk->data.r_display.expression);
}

PRIVATE tkout_t *
serv_undisplay (const tkin_t *tk)
{
	assert (tk->code == C_UNDISPLAY);
	return f1m_be_undisplay (tk->tid, tk->data.r_undisplay.dispid);
}

PRIVATE tkout_t *
serv_call (const tkin_t *tk)
{
	assert (tk->code == C_CALL);
	return f1m_be_call (tk->tid, tk->data.r_call.expression);
}

PRIVATE tkout_t *
serv_list (const tkin_t *tk)
{
	assert (tk->code == C_LIST);
	return f1m_be_list (tk->tid, tk->data.r_list.file,
					tk->data.r_list.start_line,
					tk->data.r_list.function,
					tk->data.r_list.n_lines);
}

PRIVATE tkout_t *
serv_tty (const tkin_t *tk)
{
	assert (tk->code == C_TTY);
	return f1m_be_tty (tk->tid, tk->data.r_tty.file);
}

PRIVATE tkout_t *
serv_up (const tkin_t *tk)
{
	assert (tk->code == C_UP);
	return f1m_be_up (tk->tid);
}

PRIVATE tkout_t *
serv_down (const tkin_t *tk)
{
	assert (tk->code == C_DOWN);
	return f1m_be_down (tk->tid);
}

PRIVATE tkout_t *
serv_frame (const tkin_t *tk)
{
	assert (tk->code == C_FRAME);
	return f1m_be_frame (tk->tid, tk->data.r_frame.frame);
}

PRIVATE tkout_t *
serv_info_locals (const tkin_t *tk)
{
	assert (tk->code == C_INFO_LOCALS);
	return f1m_be_info_locals (tk->tid);
}

PRIVATE tkout_t *
serv_info_args (const tkin_t *tk)
{
	assert (tk->code == C_INFO_ARGS);
	return f1m_be_info_args (tk->tid);
}

PRIVATE tkout_t *
serv_info_line (const tkin_t *tk)
{
	assert (tk->code == C_INFO_LINE);
	return f1m_be_info_line (tk->tid, tk->data.r_info_line.file,
						 tk->data.r_info_line.line,
						 tk->data.r_info_line.function);
}

PRIVATE tkout_t *
serv_info_program (const tkin_t *tk)
{
	assert (tk->code == C_INFO_PROGRAM);
	return f1m_be_info_program (tk->tid);
}

PRIVATE tkout_t *
serv_info_stack (const tkin_t *tk)
{
	assert (tk->code == C_INFO_STACK);
	return f1m_be_info_stack (tk->tid);
}

PRIVATE tkout_t *
serv_info_break (const tkin_t *tk)
{
	assert (tk->code == C_INFO_BREAK);
	return f1m_be_info_break (tk->tid);
}

PRIVATE tkout_t *
serv_info_display (const tkin_t *tk)
{
	assert (tk->code == C_INFO_DISPLAY);
	return f1m_be_info_display (tk->tid);
}

PRIVATE tkout_t *
serv_thread (const tkin_t *tk)
{
	assert (tk->code == C_THREAD);
	return f1m_be_thread (tk->tid, tk->data.r_thread.threadID);
}

PRIVATE tkout_t *
serv_info_threads (const tkin_t *tk)
{
	assert (tk->code == C_INFO_THREADS);
	return f1m_be_info_threads (tk->tid);
}

PRIVATE tkout_t *
serv_tkin (const tkin_t *tk)
{
	assert (tk->code == C_TKIN);
	return f1m_be_tkin (tk->tid, tk);
}

PRIVATE tkout_t *
serv_sendto (const tkin_t *tk)
{
	assert (tk->code == C_SENDTO);
	return f1m_be_sendto (tk->tid, tk->data.r_sendto.command);
}



/* ################################################################# */



/* -----------------------------------------------------------------
 * Function:    f1m__do_service()
 *
 * Description: Executed in a thread.  Look up in a table for
 *				appropriate function to process service 'code', and
 *				call the function.
 *
 * Returns:     NULL
 */
PRIVATE void *
f1m__do_service(void *_req)
{
	int i;
	tkout_t *tko;
	request_t *req = (request_t *)_req;
	
	for (i = 0;  serv_vect[i].funct != NULL;  i++)
		if (serv_vect[i].code == req->tk->code)
			break;

	/* Error if if not found */
	if (serv_vect[i].funct == NULL) 
		assert (0);

	/* Call the service */
	tko = serv_vect[i].funct(req->tk);
	assert (tko != NULL);

	/* Enqueue the reply */
	f1m_enqueue_reply (tko->tid, req->id, tko);

	/* Cleanup memory */
	free (req);
	return NULL;
}



/* -----------------------------------------------------------------
 * Function:    PRIVATE void *f1m__service_thread ()
 *
 * Description: Main loop of a service thread
 *
 * Returns:     NULL
 */
PRIVATE void *
f1m__service_thread (void *arg)
{
	request_t *req;
	tkout_t   *tko;
	int servid = (int)arg;
	assert (servid >= -1);
	for (;;) {
		int status;
		req = f1m_dequeue_request (servid);
		assert (req != NULL);
		assert (req->servid == servid);
		/* Execute the service in a new thread */
		status = pthread_create (&(req->thid), NULL, f1m__do_service ,req);
		assert (status == 0);
		pthread_detach (req->thid);
	}
	assert (0);					/* This point should never be reached */
	return NULL;				/* To avoid compiler warnings */
}



PRIVATE int
f1m__start_service_thread (int servid)
{
	int err;
	pthread_t thid;

	/* Be sure the service */
	assert (servid >= -1);
	
	/* Create a new 'serv_thread_t' structure and fill it */
	err = pthread_create (&thid, pthread_attr_default,
						  f1m__service_thread, (void *)servid);
	assert (err == 0);
	return err;
}




/* #################################################################
 *		PUBLIC functions
 * ################################################################# */
 
/* -----------------------------------------------------------------
 * Function:    int status = f1m_be_initialize ()
 *
 * Description: Start the system thread
 *
 * Returns:     The status
 */
PUBLIC int
f1m_be_initialize (void)
{
	f1m_servdata_initialize ();

	/* Wait for client connections */
	return f1m__start_service_thread (NO_SERVER);
}


/* -----------------------------------------------------------------
 * Function:    int status = f1m_be_terminate_1 ()
 *
 * Description: Terminate the system thread
 *
 * Returns:     The status
 */
PUBLIC int
f1m_be_terminate_1 (int dummy)
{
	/* Terminate threads */
	f1m_servdata_terminate ();
	return 0;
}

