/* *****************************************************************
 *
 *  $Id: f1m_lib.c,v 1.4 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/16 19:10:16 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
  *  Description: Implementation of the services provided by FIDDLE-1
 *
 *		f1m_initialize ()		-- Initialize FIDDLE-1
 *		f1m_terminate ()		-- Terminate FIDDLE-1
 *		f1m_tids ()				-- List existing task IDs
 *		f1m_kill ()				-- Kill a target process
 *		f1m_add_debugger ()		-- Add a component-level-debugger
 *		f1m_attach ()			-- Attach to a running process
 *		f1m_detach ()			-- Detach from a running process
 *		f1m_symbol_file ()		-- Specify whre to find program symbols
 *		f1m_file ()				-- Load program and symbols into memory
 *		f1m_run ()				-- Run a program
 *		f1m_step ()				-- Executite until the next instruction
 *		f1m_next ()				-- Executite until the next instruction
 *		f1m_continue ()			-- Continue the execution
 *		f1m_finish ()			-- Execute until returning from
 *								   current stack frame
 *		f1m_signal ()			-- Send a signal to a process
 *		f1m_break ()			-- Set a breakpoint
 *		f1m_delete ()			-- Delete a breakpoint
 *		f1m_set_variable ()		-- Change the contents of a variable
 *		f1m_call ()				-- Call a function in the current
 *								   context of the target process
 *		f1m_evaluate ()			-- Evaluate an expression in the current
 *								   context of the target process
 *		f1m_display ()			-- Evaluate and display  an expression 
 *								   every time the process execution stops
 *		f1m_undisplay ()		-- Undisplay an expression
 *		f1m_list ()				-- List the process source code
 *		f1m_info_locals ()		-- Get name and value of local variables
 *		f1m_info_args ()		-- Get name and value of function argument
 *		f1m_info_line ()			-- Get info about current line
 *		f1m_info_program ()		-- Get info about current program
 *		f1m_info_stack ()		-- Get info about current stack frames
 *		f1m_up ()				-- Go up in the stack frame list
 *		f1m_down ()				-- Go down in the stack frame list
 *		f1m_frame ()			-- Select a specific stack frame
 *		f1m_info_break ()		-- Get info about current breakpoint
 *		f1m_info_display ()		-- Get info about current display expressions
 *		f1m_thread ()			-- Select a thread
 *		f1m_info_threads ()		-- Get info about the existing threads
 *		f1m_tty ()				-- Set a TTY for future IO
 *		f1m_sendto ()			-- Send a command directly to the debugger
 *
 * -----------------------------------------------------------------
 */

#include "config.h"
#include "glob.h"

#include <stdlib.h>
#include <assert.h>
#include "f0s_tkout.h"
#include "f0s_out_codes.h"

#include "f0m_xdr.h"
#include "f0m_tkin.h"

#include "f1m_queues.h"
#include "f1m_back_end.h"
#include "f1m_lib.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


 
/* -----------------------------------------------------------------
 * Function:    f1m_initialize()
 *
 * Description: Initialize all the necessary data structures
 *
 * Returns:     The status
 */
PUBLIC int f1m_initialize (void)
{	
	/* Initialize Queues */
	f1m_queues_init ();
	
	/* Start system thread */
	f1m_be_initialize ();

	return 0;
}


/* -----------------------------------------------------------------
 * Function:    f1m_terminate_1()
 *
 * Description: Terminate the debugging in this node.  Detach from all 
 *				processes and kil all CLDs.
 *
 * Returns:     The status
 */
PUBLIC int f1m_terminate_1 (int halt)
{
	/* Terminate system thread */
	f1m_be_terminate_1 (halt);
	
	/* Terminate Queues */
	f1m_queues_terminate ();

	/* Terminate lower layer */
	return f0m_terminate ();
}


/* -----------------------------------------------------------------
 * Function:    f1m_tids()
 *
 * Description: Get a list of all existing process TIDs.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_tids (const char *options)
{
	tkin_t *req = f0m_tkin_new (NO_SERVER, C_TIDS, options);
    return f1m_tkin3 (NO_SERVER, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_add_debugger ()
 *
 * Description: Add a 'ready to be controlled' debugger to Fiddle
 *				'dbg_pid' should contain the PID of the debugger
 *				'tp_pid' should contain the PID of the target process
 *						 or -1 if process is not loaded or not running yet
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_add_debugger (const char *machine, int dbg_pid, int tp_pid,
						   const char *cmd_channel, const char *rep_channel)
{
	tkin_t *req = f0m_tkin_new (NO_SERVER, C_ADD_DEBUGGER, machine,
								dbg_pid, tp_pid, cmd_channel, rep_channel);
    return f1m_tkin3 (NO_SERVER, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_kill()
 *
 * Description: Kill a process being debugger but let the CLD
 *				    available for a next usage
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_kill (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_KILL);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_attach()
 *
 * Description: Find an unsused component-level debugger or start one
 *				if none was found.  Then attach the debugger to the
 *				process 'tp_pid'.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_attach (const char *machine, int tp_pid)
{
	tkin_t *req = f0m_tkin_new (NO_SERVER, C_ATTACH, tp_pid, machine);
    return f1m_tkin3 (NO_SERVER, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_detach()
 *
 * Description: Detach a component-level debugger from the process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_detach (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_DETACH);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_symbol_file()
 *
 * Description: Load the symbols
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_symbol_file (int tid, const char *machine, const char *file)
{
	tkin_t *req = f0m_tkin_new (tid, C_SYMBOL_FILE, file, machine);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_file()
 *
 * Description: Load a process image into memory
 *				'tid == -1' load a new process image
 *				'tid != -1' reload a process image
 *				'tid' and 'machine' are exclusive
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_file (int tid, const char *machine, const char *file)
{
	tkin_t *req;

	if (tid == -1)
		tid = NO_SERVER;
	req = f0m_tkin_new (tid, C_FILE, file, machine);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_run()
 *
 * Description: Start the execution of a process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_run (int tid, const char *args)
{
	tkin_t *req = f0m_tkin_new (tid, C_RUN, args);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_step()
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps into the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_step (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_STEP);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_next()
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps over the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_next (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_NEXT);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_continue()
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_continue (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_CONTINUE);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_finish()
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_finish (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_FINISH);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_signal()
 *
 * Description: Send a signal to a process.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_signal (int tid, int signum)
{
       tkin_t *req;

       req = f0m_tkin_new (tid, C_SIGNAL, tid, signum);
	   return f1m_tkin3 (NO_SERVER, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_break()
 *
 * Description: Set a breakpoint.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_break (int tid, const char *file, int line, const char *function)
{
	tkin_t *req = f0m_tkin_new (tid, C_BREAK, file, line, function);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_delete()
 *
 * Description: Remove a breakpoint.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_delete (int tid, int bpid)
{
	tkin_t *req = f0m_tkin_new (tid, C_DELETE, bpid);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_set_variable()
 *
 * Description: Set 'variable' to 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_set_variable (int tid, const char *variable,
						  const char *expression)
{
	tkin_t *req = f0m_tkin_new (tid, C_SET_VARIABLE, variable, expression);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_evaluate()
 *
 * Description: Evaluate an 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_evaluate (int tid, const char *expression)
{
	tkin_t *req = f0m_tkin_new (tid, C_EVALUATE, expression);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_display()
 *
 * Description: Display 'expression' every time the program stops
 *				If 'expression == NULL' list list all expressions/values
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_display (int tid, const char *expression)
{
	tkin_t *req = f0m_tkin_new (tid, C_DISPLAY, expression);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_undisplay()
 *
 * Description: Remove a display expression
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_undisplay (int tid, int dispid)
{
	tkin_t *req = f0m_tkin_new (tid, C_UNDISPLAY, dispid);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:	tkof1m_call()
 *
 * Description: Call a function and get it's return value
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_call (int tid, const char *expression)
{
	tkin_t *req = f0m_tkin_new (tid, C_CALL, expression);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_list()
 *
 * Description: List a source file.  
 *              The arguments 'start_line' and 'function' are exclusive.
 *              The arguments 'n_lines' is optional.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_list (int tid, const char *file, int start_line,
				 const char *function, int n_lines)
{
	tkin_t *req = f0m_tkin_new (tid, C_LIST, file, start_line, function, n_lines);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:	tkof1m_tty()
 *
 * Description: Set the I/O tty for the process being debugged
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_tty (int tid, const char *file)
{
	tkin_t *req = f0m_tkin_new (tid, C_TTY, file);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_locals()
 *
 * Description: Get name and value of local variables
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_locals (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_LOCALS);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_args()
 *
 * Description: Get name and value of current function arguments
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_args (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_ARGS);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_line()
 *
 * Description: Get information about a source line.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_line (int tid, const char *file, int line,
					  const char *function)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_LINE, file, line, function);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:	tkof1m_info_program()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_program (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_PROGRAM, req);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_stack()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_stack (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_STACK);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_up()
 *
 * Description: Go "up" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_up (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_UP);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_down()
 *
 * Description: Go "down" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_down (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_DOWN);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_frame()
 *
 * Description: Select a specific stack frame
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_frame (int tid, int frame)
{
	tkin_t *req = f0m_tkin_new (tid, C_FRAME, frame);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_break()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_break (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_BREAK);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_display()
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_display (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_DISPLAY);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_thread()
 *
 * Description: Switch the current thread.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_thread (int tid, int threadID)
{
	tkin_t *req = f0m_tkin_new (tid, C_THREAD, threadID);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_info_threads()
 *
 * Description: Get information about the existing threads.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_info_threads (int tid)
{
	tkin_t *req = f0m_tkin_new (tid, C_INFO_THREADS);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_sendto()
 *
 * Description: Send a command string directly to the debugger.
 *				'tkout' contains the output (NOT PARSED).
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_sendto (int tid, const char *command)
{
	tkin_t *req = f0m_tkin_new (tid, C_SENDTO, command);
    return f1m_tkin3 (tid, req, 1);
}


/* -----------------------------------------------------------------
 * Function:    f1m_tkin3()
 *
 * Description: Set a command already packed into a 'tkin_t' strucure
 *				to process 'tid'.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f1m_tkin3 (int tid, tkin_t *tk, int delete)
{
	int reqid;
	tkout_t *rep;

	assert (tk != NULL);
	reqid = f1m_enqueue_request (tid, tk);
	if (reqid >= 0)
		rep = f1m_dequeue_reply (reqid);
	else
		rep = f0s_tkout_new (E_RTS_INVALID_TID);
	if (delete)
		f0m_tkin_delete (tk);
	return rep;
}
