/* *****************************************************************
 *
 *  $Id: f1m_servdata.h,v 1.2 2002/07/07 22:38:34 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:44:41 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Mantain a database that maps 'Global Task IDs' into
 *				     'Local Task IDs' and the I/O channels to the
 *				     associated F1 server.
 *
 *		f1m_servdata_initialize ()	-- Initialize the database
 *		f1m_servdata_terminate ()    -- Terminate the database
 *		f1m_gtid_2_servdata ()		-- Get data from database
 *		f1m_add_servdata ()		    -- Get data from database
 *
 * -----------------------------------------------------------------
 */
#ifndef __F1M_SERVDATA_H
#define __F1M_SERVDATA_H

#include "glob.h"

#include <pthread.h>

#include "f0s_lib.h"
#include "f0m_tkin.h"

__BEGIN_DECLS


typedef struct serv_data_machine_s serv_data_machine_t;
struct serv_data_machine_s {
	int   servid;
	char *machine;
	int   fd[2];
	pthread_t thid;				/* to read the tkout's */
	chain_t *p_reps;			/* pending TkOut's from this server */
	pthread_mutex_t notid_mutex; /* used when the tid sent is -1 */
};


typedef struct serv_data_tid_s serv_data_tid_t;
struct serv_data_tid_s {
	int gtid;					/* Global Task ID */
	int servid;					/* The server ID */
	int ltid;					/* Local Task ID */
	int fd[2];					/* READ and WRITE file descriptors */
};


typedef struct client_data_s client_data_t;
struct client_data_s {
	int pid;					/* Process ID */
	char *filename;				/* Name of client executable */
	char *hostaddr;				/* Host address where client is running */
};





/* -----------------------------------------------------------------
 * Function:    f1m_fc_clients ()
 *
 * Description: Return a pointer to a chain with client data
 *
 * Returns:     Pointer to the chain with data or NULL on error
 */
PUBLIC chain_t *
f1m_fc_clients (void);



/* -----------------------------------------------------------------
 * Function:    f1m_fc_add_client_data ()
 *
 * Description: Add data to the 'clients' database
 *
 * Returns:     Pointer to the node with data or NULL on error
 */
PUBLIC int
f1m_fc_add_client_data (int pid, const char *filename, const char *hostaddr);


/* -----------------------------------------------------------------
 * Function:    f1m_fc_remove_client_data ()
 *
 * Description: Remove data from the 'clients' database
 *
 * Returns:     Number of remaining clients
 */
PUBLIC int
f1m_fc_remove_client_data (int pid);


/* -----------------------------------------------------------------
 * Function:    int f1m_servdata_initialize (void)
 *
 * Description: Initialize the server information database
 *
 * Returns:     0
 */

int f1m_servdata_initialize (void);


/* -----------------------------------------------------------------
 * Function:    int f1m_servdata_terminate (void)
 *
 * Description: Terminate the server information database
 *
 * Returns:     0
 */
int f1m_servdata_terminate (void);



/* -----------------------------------------------------------------
 * Function:    serv_data_tid_t *f1m_add_servdata_tid (int gtid,
 *											          int servid, int ltid, 
 *													  int rfd, int wfd)
 *
 * Description: Add data to the 'tids' database
 *
 * Returns:     Pointer to the node with data or NULL on error
 */
serv_data_tid_t *f1m_tp_add_servdata_tid (int gtid, int servid, int ltid,
									  int rfd, int wfd);



/* -----------------------------------------------------------------
 * Function:    serv_data_tid_t *f1m_sltid_2_tiddata (int servid, int ltid)
 *
 * Description: Get the server data ssociated with a pair (serverid, ltid)
 *
 * Returns:     Pointer to data
 */
serv_data_tid_t *f1m_tp_sltid_2_tiddata (int servid, int ltid);


/* -----------------------------------------------------------------
 * Function:    serv_data_machine_t *f1m_add_servdata_machine (
 *												int servid,
 *												const char *machine, 
 *											    int rfd, int wfd)
 *
 * Description: Add data to the 'machines' database
 *
 * Returns:     Pointer to the node with data or NULL on error
 */
serv_data_machine_t *f1m_tp_add_servdata_machine (int servid,
											  const char *machine,
											  int rfd, int wfd);


/* -----------------------------------------------------------------
 * Function:    serv_data_t *f1m_gtid_2_servdata (int gtid)
 *
 * Description: Get the server data ssociated with a global 'gtid' 
 *
 * Returns:     Pointer to data
 */
serv_data_tid_t *f1m_tp_gtid_2_servdata (int gtid);


/* -----------------------------------------------------------------
 * Function:    serv_data_machine_t *f1m_machine_2_servdata (const char *machine)
 *
 * Description: Get the server data ssociated with a machine 
 *
 * Returns:     Pointer to data
 */
serv_data_machine_t *f1m_tp_machine_2_servdata (const char *machine);


/* -----------------------------------------------------------------
 * Function:    serv_data_machine_t *f1m_servid_2_servdata (int servid)
 *
 * Description: Get the server data ssociated with a server ID 
 *
 * Returns:     Pointer to data
 */
serv_data_machine_t *f1m_tp_servid_2_servdata (int servid);


/* -----------------------------------------------------------------
 * Function:    chain_t *f1m_apply_to_all_machines (tkin_t *req)
 *
 * Description: Request a service to all the machines
 *
 * Returns:     List of replies (tkout_t)
 */
chain_t *f1m_tp_apply_to_all_machines (tkin_t *req);


/* -----------------------------------------------------------------
 * Function:    tkout_t *f1m_read_tkout (int servid, int ltid)
 *
 * Description: Read a tkout
 *
 * Returns:     The tkout or NULL on error
 */
tkout_t *f1m_read_tkout (int servid, int ltid);

__END_DECLS


#endif /* __F1M_SERVDATA_H */
