/* *****************************************************************
 *
 *  $Id: f1m_serv.c,v 1.10 2002/09/27 13:31:51 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:47:03 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */


#include "config.h"
#include "glob.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <netdb.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#include <pthread.h>

#include "f0s_tkout.h"

#include "f0m_tkin.h"
#include "f0m_rw_to_file.h"

#include "f1m_servdata.h"
#include "f1m_lib.h"

#include "sock_ntop.h"




/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#define LISTENQ 10

#define VERBOSE(exp) if (options.verbose) printf exp


typedef struct options_s options_t;
struct options_s {
    int debug;					/* 0 */
    int keep;					/* 0 */
    int verbose;				/* 0 */
    char *service;				/* F1M_SERVER_SERVICE_NUMBER */
    char *xdr;					/* NULL */
};


options_t options = {0, 0, 0, F1M_SERVER_SERVICE_NUMBER, NULL};


typedef struct client_data {
    int fd;
    pthread_t thid;
    const char *hostaddr;
} sock_data_t;


#define TERM_QUIT   0
#define TERM_HALT	1


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    int f1mserv__update_client_counter ()
 *
 * Description: Updates the counter associated with the number of
 *				active clients
 *
 * Returns:     The number of active clients
 */
PRIVATE int
f1mserv__update_client_counter (int val)
{
    static int active = 0;
    static pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;
    int local_active;

    pthread_mutex_lock (&m);
    active += val;
    local_active = active;
    pthread_mutex_unlock (&m);
    return local_active;
}



/* -----------------------------------------------------------------
 * Function:    void f1mserv__terminate ()
 *
 * Description: f1mserv__Terminate the f1m_server
 *
 * Returns:     Never returns!
 */
PRIVATE void
f1mserv__terminate (int kind, int cli_pid)
{
	if (cli_pid != -1)
		f1m_fc_remove_client_data (cli_pid);
    if (kind == TERM_HALT ||
        (f1mserv__update_client_counter (-1) == 0 && options.keep == 0)) {
        /* No more clients. f1mserv__Terminate the server */
        f1m_terminate ();
        exit (0);
    }
    else
        /* Other clients still active. Just f1mserv__terminate this thread. */
        pthread_exit (NULL);
}



/* -----------------------------------------------------------------
 * Function:    int f1mserv__tcp_listen ()
 *
 * Description: Create a TCP socket, bind the server's port, and allow 
 *				incomming connection requests to be accepted
 *				(adapted from "Unix Networking Programming", Vol. I, 
 *				 by W. Richard Stevens)
 *
 * Returns:     
 */
PRIVATE int
f1mserv__tcp_listen (const char *host, const char *serv, socklen_t * addrlenp)
{
    const int on = 1;
    int listenfd, n;
    struct addrinfo hints, *res, *ressave;

    memset (&hints, 0, sizeof (hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_INET;
    hints.ai_socktype = SOCK_STREAM;
    n = getaddrinfo (host, serv, &hints, &res);
    assert (n == 0);

    for (ressave = res; res != NULL; res = res->ai_next) {
        listenfd = socket (res->ai_family,
                           res->ai_socktype, res->ai_protocol);
        if (listenfd > 0) {
            setsockopt (listenfd, SOL_SOCKET, SO_REUSEADDR,
                        &on, sizeof (on));
            if (bind (listenfd, res->ai_addr, res->ai_addrlen) == 0)
                break;          /* Success!!! */
            close (listenfd);
        }
    }
    assert (res != NULL);
    listen (listenfd, LISTENQ);
    if (addrlenp)
        *addrlenp = res->ai_addrlen;
    freeaddrinfo (ressave);
    return listenfd;
}




/* -----------------------------------------------------------------
 * Function:    void f1mserv__usage ()
 *
 * Description: Print the f1mserv__usage string and exits
 *
 * Returns:     Never returns!!!
 */
PRIVATE void
f1mserv__usage (const char *progname)
{
    printf ("\n\
f1mserv__Usage: %s [-h|-?] [-d] [-v] [-k] [-s <service/port#>] -x [jml|xml]\n\
\n\
Option: -h | -?             Print this help message.\n\
        -d                  Print debug information.\n\
        -v                  Be verbose.\n\
        -k                  Keep server alive.\n\
        -s <service/port#>  Service name / port number.\n\
        -x [jml|xml]        XDR protocol used.\n\
\n\
", progname);
    exit (1);
}



/* -----------------------------------------------------------------
 * Function:    void f1mserv__sig_handler ()
 *
 * Description: Signal handler
 *
 * Returns:     
 */
PRIVATE void
f1mserv__sig_handler (int signum)
{
    VERBOSE (("Signal %d received!  Exiting...\n", signum));
    f1mserv__terminate (TERM_HALT, -1);
}



/* -----------------------------------------------------------------
 * Function:    int f1mserv__check_options ()
 *
 * Description: Check and validate the command line options
 *
 * Returns:     
 */
PRIVATE int
f1mserv__check_options (int argc, char *argv[])
{

/* OPTION:	s	The general service name / service number */
#define THE_OPITIONS "h?dvks:x:"

    int ch;

    while ((ch = getopt (argc, argv, THE_OPITIONS)) != EOF) {
/* 		assert (ch != '?'); */
        switch (ch) {
        case 'h':
        case '?':
            f1mserv__usage (argv[0]);
            break;
        case 'd':
            options.debug = 1;
            break;
        case 'k':
            options.keep = 1;
            break;
        case 'v':
            options.verbose = 1;
            break;
        case 's':
            options.service = optarg;
            break;
         case 'x':
            options.xdr = optarg;
			setenv ("FIDDLE_XDR", optarg, 1);
            break;
		}
	}
    return optind;
}


/* -----------------------------------------------------------------
 * Function:    void f1mserv__filter_clients ()
 *
 * Description: Handler to "chain_prune()", to eliminate unwanted
 *				clients from list.
 *
 * Returns:     Never returns!!!  BUG!
 */
int f1mserv__filter_clients (void *data, void *args)
{
	rclients_t *cli = (rclients_t *)data;
	const char *fm = (const char *)args;
	const char *p;

	assert (fm != NULL);
	switch (*(fm+1)) {
	case 'f':
		p = cli->filename;
		break;
	case 'm':
		p = cli->hostaddr;
		break;
	default:
		return 0;				/* Do nothing */
	}
	if (strncmp	(p, fm+3, strlen (fm+3)) == 0)
		return 0;
	else return 1;
}


/* -----------------------------------------------------------------
 * Function:    f1mserv__get_client_registration()
 *
 * Description: Get a tkin from input descriptor, which must be a
 *				C_REGISTER_CLIENT.  Reply to the client with OK token.
 *
 * Returns:     The client pid (extracted from the input token)
 */
PRIVATE int
f1mserv__get_client_registration (const sock_data_t *sd)
{
	tkin_t  *tkin;
	tkout_t *tkout;
	int cli_pid;
	int err;

	assert (sd != NULL);
	assert (sd->fd >= 0);
	
	tkin = f0m_read_tkin (sd->fd);
	if (tkin == NULL) return -1; /* EOF before registration */
	assert (tkin->code == C_REGISTER_CLIENT);

	/* Get the client pid from tkin */
	cli_pid = tkin->data.r_register_client.pid;

	/* And register in database */
	f1m_fc_add_client_data (cli_pid,
							tkin->data.r_register_client.filename,
							sd->hostaddr);

	/* Send the reply to client */
	tkout = f0s_tkout_new (FIDDLE_OK);  /* No error */
	err = f0m_write_tkout (sd->fd, tkout);
	assert (err >= 0);
	f0s_tkout_delete (tkout);

	/* Return the extracted client PID */
	return cli_pid;
}


typedef struct tkin_and_fd_s tkin_and_fd_t;
struct tkin_and_fd_s {
	tkin_t		*tkin;
	int			 fd;
	pthread_t	 thid;
};



/* -----------------------------------------------------------------
 * Function:    f1mserv__process_service_requested()
 *
 * Description: Thread to process a single service request.
 *
 * Returns:     The client pid (extracted from the input token)
 */
PRIVATE void *
f1mserv__process_service_requested (void *_tk)
{
	int err;
	tkout_t *tkout;
	tkin_and_fd_t *tk = (tkin_and_fd_t *)_tk;
	
	assert (tk != NULL);
	assert (tk->tkin != NULL);
	assert (tk->fd >= 0);
	
	/* Send the request to layer (1m) and get the reply */
	tkout = f1m_tkin3 (tk->tkin->tid, tk->tkin, 1);
	assert (tkout != NULL);

	/* Send the reply (with exclusive access to STDOUT) */
	err = f0m_write_tkout (tk->fd, tkout);
	assert (err >= 0);

	/* Destroy the output token */
	f0s_tkout_delete (tkout);
	free (tk);

	return NULL;
}


/* -----------------------------------------------------------------
 * Function:    f1mserv__service_thread()
 *
 * Description: Main loop of the thread created to attend requests
 *				from one client
 *
 * Returns:     Never returns!!!  BUG!
 */
PRIVATE void *
f1mserv__service_thread (void *_sd)
{
    int cli_pid;
    int err;
    tkin_t *tkin;
    tkout_t *tkout;
    int term_code;


    sock_data_t *sd = (sock_data_t *) _sd;
    assert (sd->fd >= 0);

	/* One more client is active */
    f1mserv__update_client_counter (1);

	/* First message from client must be the client registration */

	cli_pid = f1mserv__get_client_registration (sd);

	/* If registration was successful */
	if (cli_pid != -1) {
		term_code = -1;
		while (term_code == -1) {
			/* Get the 'tkin_t' structure */
			tkin = f0m_read_tkin (sd->fd);
			if (tkin == NULL)
				break;
			
			switch (tkin->code) {
			case C_TERM_SERVER:    /* Kill server */
				tkout = f0s_tkout_new (E_RTS_OK);
				assert (tkout != NULL);
				term_code = tkin->data.r_terminate.halt;
				break;
			case C_CLIENTS:        /* Return chain with fiddle clients */
				tkout = f0s_tkout_new (E_RTS_OK);   /* No error */
				assert (tkout != NULL);
				tkout->rep = f0s_tkout_data_new ();
				tkout->rep->code = R_CLIENTS;
				tkout->rep->str = NULL;
				/* Assume no arguments were given */
				chain_merge (tkout->rep->data, f1m_fc_clients ());
				
				if (tkin->data.r_tids_clients.options != NULL
					&&  *(tkin->data.r_tids_clients.options) != '\0') {
					/* There are some argumetns... test "-f" and "-m"  */
					char *p;	/* p will point to NULL or to the
								   filename, then p will point to NULL
								   or to the machine name */
					p = strstr (tkin->data.r_tids_clients.options, "-f");
					if (p != NULL)
						chain_prune (tkout->rep->data, f1mserv__filter_clients, p);
					p = strstr (tkin->data.r_tids_clients.options, "-m");
					if (p != NULL)
						chain_prune (tkout->rep->data, f1mserv__filter_clients, p);
				}
				break;
			case C_REGISTER_CLIENT:    /* Register client in database */
				assert (0);			/* Only one registration is accepted,
									   which must be done before the main
									   loop */
				break;
			default:               /* Request the service */
				{
					tkin_and_fd_t *tk = malloc (sizeof (*tk));
					assert (tk != NULL);
					tk->fd = sd->fd;
					tk->tkin = tkin;
					pthread_create (&(tk->thid), NULL,
									f1mserv__process_service_requested, tk);
/* 					f1mserv__process_service_requested (tk); */
					pthread_detach (tk->thid);
					continue;			/* Ugly solution!  :-( */
				}
			}
			
			/* Send the reply (with exclusive access to STDOUT) */
			err = f0m_write_tkout (sd->fd, tkout);
			assert (err >= 0);

			/* Destroy the output token */
			f0s_tkout_delete (tkout);
		}
    }
	free (sd);
    f1mserv__terminate (term_code, cli_pid);
    return NULL;
}




/* -----------------------------------------------------------------
 * Function:    void f1mserv__setup_server ();
 *
 * Description: Setup a server listening in <service>
 *
 * Returns:     Never returns!!!
 */
PRIVATE void
f1mserv__setup_server (const char *service)
{
    int listenfd;
    socklen_t addrlen, len;
    struct sockaddr *cliaddr;

    listenfd = f1mserv__tcp_listen (NULL, service, &addrlen);
    assert (listenfd >= 0);
    assert (addrlen > 0);
    cliaddr = alloca (addrlen);

    for (;;) {
        sock_data_t *sd = malloc (sizeof (*sd));
        len = addrlen;
        sd->fd = accept (listenfd, cliaddr, &len);
		assert (sd->fd >= 0);
        sd->hostaddr = sock_ntop (cliaddr, len);
        VERBOSE (("Connection from %s\n", sd->hostaddr));
        pthread_create (&(sd->thid), NULL, f1mserv__service_thread, sd);
    }
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */




/* -----------------------------------------------------------------
 * Function:    int main ()
 *
 * Description: 
 *
 * Returns:     
 */
int
main (int argc, char *argv[])
{
    int optind = f1mserv__check_options (argc, argv);

    VERBOSE (("SERVICE NAME/PORT# = %s\n", options.service));

    if (optind != argc)
        f1mserv__usage (argv[0]);   /* Calls 'exit (1)' */

    if (signal (SIGINT, f1mserv__sig_handler) == SIG_ERR)
        assert (0);

    f1m_initialize ();

    f1mserv__setup_server (options.service);
    return 0;                   /* To avoid warinings with "-Wall" */
}
