/* *****************************************************************
 *
 *  $Id: fx.c,v 1.4 2002/09/28 00:20:27 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/27 17:03:53 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */



#define TEST_ERROR(e)		test_error(e)
#define TEST_REPLY(e)		test_reply(e)
#define TEST_OOB(e)		test_oob(e)
#define TEST_POSITION(e)	test_position(e)



#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>
#include <pthread.h>
#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include "chain.h"

#include "fx_lib.h"

#include "test_output.h"
#include "cmds.h"
#include "error.h"
#include "timer.h"


typedef struct options_s options_t;
struct options_s {
	int			time;			/* TRUE for "-t" */
	char	   *input;			/* Input define "-i dev" */
	char	   *output;			/* Output define "-o dev" */
	chain_t	   *file;			/* Files to load */
	int			pid;			/* PID of target process */
#ifdef f2m
	char	   *server;			/* Where is "f1m_serv" running */
	char	   *port;			/* What port to connect */
	int			flags;			/* 0 | START_SERVER | KEEP_SERVER */
#endif
};


options_t options;

extern char default_tid[];


void console_error (int status)
{
	const char *msg = cons_err_msg (status);
	if (msg != NULL)
		printf ("\nConsole Error %d: %s\n", status, msg);
}






/* -----------------------------------------------------------------
 * Function:    void usage (const char *progname)
 *
 * Description: Print the usage string and exits
 *
 * Returns:     Never returns!!!
 */
PRIVATE void usage (const char *progname)
{
#ifdef f2m	
	char buf[256];
	int n;
#endif
	
	printf ("\n\n\
Usage: %s [-h|-?] [-t] [-i tty] [-o tty] [-x jml|xml]\n\
", progname);
#ifdef f2m
	sprintf (buf, "Usage: %s ", progname);
	n = strlen (buf);
	while (n--)
		printf (" ");
	printf ("[-r [-k]] [-s <name/IP>] [-p <service/port#>]");
#endif
	printf ("\n\n\
Option: -h | -?             Print this help message.\n\
        -t                  Print time (performance) information.\n\
        -i tty              Get console input from <tty>.\n\
        -o tty              Send console output to <tty>.");
#ifdef f2m
	printf ("\n\
        -r                  Start a new server if not found.\n\
        -k                  Keep server alive (only if a new one is started).\n\
        -s <name/IP>        Name / IP address of server.\n\
        -p <service/port#>  Service name / port number.");
#endif
	printf ("\n\
        -a <pid>            Attach to process <pid>.\n\
        -x <jml|xml>        XDR protocol to use.");
	puts ("\n\n");
	exit (1);
}



/* -----------------------------------------------------------------
 * Function:    char *cmdstr = command_completion (IN const char *text, 
 *														 IN int state)
 *
 * Description: Complete a command for <TAB> expansion using the
 *				readline library.
 *
 * Returns:     A pointer to the copy of the possible expansion.
 */
PRIVATE char *command_completion (const char *text, int state)
{
    PRIVATE int locstate;
    
    if (state == 0)
		locstate = 0;
    else
		locstate++;
    while (cmd_table[locstate].cmd_funct != 0) {
		if (strstr (cmd_table[locstate].cmd_str,
					text) == cmd_table[locstate].cmd_str)
			return strdup (cmd_table[locstate].cmd_str);
		else
			locstate++;
    }
    return 0;
}



/* -----------------------------------------------------------------
 * Function:    const DbgCommand *cmd = find_command (IN const char *cmd)
 *
 * Description: Find a command structure in the table
 *
 * Returns:     
 */
PRIVATE const DbgCommand *find_command (const char *comm)
{
	char *exp[256];
    int i;
    for (i = 0;  cmd_table[i].cmd_funct != 0;  i++)
		if (strcmp (comm, cmd_table[i].cmd_str) == 0)
			break;

	/* If not found, try to expand */
	if (cmd_table[i].cmd_funct == 0) {
		int j;
		for (j = 0;  (exp[j] = (char *)command_completion (comm, j)); j++)
			;
		if (j > 1) {
			int k;
			printf ("Possible expansions are: ");
			for (k = 0;  k < j;  k++)
				printf ("\"%s\" ", exp[k]);
			puts ("");
		}
		else if (j == 1) {
			printf ("Accepted as \"%s\"\n", exp[0]);
			for (i = 0;  cmd_table[i].cmd_funct != 0;  i++)
				if (strcmp (exp[0], cmd_table[i].cmd_str) == 0)
					break;
		}
		for (;  j;  j--)
			free (exp[j]);
	}

	/* Return whatever */
    return &cmd_table[i];
}




/* ================================================================= */
/* ================================================================= */
/* ================================================================= */
/* ================================================================= */


/* ================================================================= */
/* ================================================================= */
/* ================================================================= */
/* ================================================================= */

#ifdef HAVE_LIBREADLINE
PRIVATE char *rl_gets (const char *level)
{
	PRIVATE char prompt[256];
	PRIVATE char *line_read = 0;

	sprintf (prompt, "%s [%s]> ", level, default_tid);

	do {
		puts ("");
		if (line_read) {
			free (line_read);
			line_read = 0;
		}
		line_read = readline (prompt);
	} while (line_read  &&  strlen (line_read) == 0);
	if (line_read)
		add_history (line_read);
    return line_read;
}
#else
PRIVATE char *rl_gets (const char *level)
{
#define MAXLINELEN 512
	PRIVATE char prompt[256];
	PRIVATE char *line_read = 0;
	int len;
	char line[MAXLINELEN];
	
		sprintf (prompt, "%s [%s]> ", level, default_tid);

	do {
		puts ("");
		printf (prompt); fflush (stdin);
		fgets (line, MAXLINELEN, stdin);
		len = strlen (line);
		if (line[len-1] == '\n') {	/* Remove trailing '\n' if present */
			line[len-1] = '\0';
			len --;
		}
	} while (len == 0);
	if (line_read)
		free (line_read);
	line_read = strdup (line);
    return line_read;
}

#endif



#define MAX_ARGV 32

typedef struct func_data_s func_data_t;
struct func_data_s {
	int				background;	/* False if must wait */
	CmdFunction_f  *funct;		/* The function to call */
	pthread_t       th_id;		/* The id of the executing thread */
	int				argc;		/* Number of arguments */
	const char	   *argv[MAX_ARGV];	/* The cmd line arguments */
};

PRIVATE int execute_command (const char *oline);

PRIVATE void *execute_command_thread (void *args)
{
	func_data_t *data;
	tkout_t *tkout;
	Time elapsed;

	/* Execute the command */
	assert (args != NULL);
	data = (func_data_t *)args;
	if (options.time)
		start_timer ();
	tkout = data->funct (data->argc, data->argv);
	assert (tkout != NULL);
	if (options.time)
		elapsed = stop_timer ();

	/* Check the reply */
	if (tkout->status < FIDDLE_OK) 
		console_error (tkout->status);
	else if (tkout->status > FIDDLE_OK)
		was_error (tkout->status, tkout->str);
	else if (tkout->err != NULL)
		TEST_ERROR (tkout->err);
	else {
		int print_line = 0;
		int tid = 0;
		/* No error detected */
		if (options.time)
			print_time (data->argv[1], elapsed);
		TEST_REPLY (tkout->rep);
		TEST_OOB (tkout->oob);
		TEST_POSITION (tkout->pos);

		/* Print current line if position information */
		if (tkout->pos != NULL) {
			print_line = ((pposition_t *)chain_first_data (tkout->pos->data))->line;
			tid = tkout->tid;
		}
		if (print_line) {
			char newline[256];
			sprintf (newline, "%d list %d 1", tid, print_line);
			execute_command (newline);
		}
	}
	f0s_tkout_delete (tkout);
/* 	free (data->argv); */
/* 	free (data); */
	return NULL;
}


PRIVATE void execute_command_function (func_data_t *data)
{
	pthread_attr_t attr;

	pthread_attr_init (&attr);
	if (data->background)
		pthread_attr_setdetachstate (&attr, 1);
	pthread_create (&data->th_id, &attr, execute_command_thread, data);
	if (!data->background)
		pthread_join (data->th_id, NULL);
}

/* -----------------------------------------------------------------
 * Function:    void execute (IN const char *command)
 *
 * Description: Execute a command
 *
 * Returns:     The status of the operation
 */
PRIVATE void execute (const char *command)
{
	/* Define delimiters as <SAPCE> and <TAB> */
#define DELIM " \t"
	/* Use strtok_r() if possible, strtok() otherwise */
#ifdef __USE_POSIX
	char *save;
#else
#define strtok_r(s,d,x) strtok(s,d)
#endif
	const DbgCommand *cmdstruct;
	func_data_t *data;
	const char *lastarg;
	
	/* Allocate memory for data  */
	data = malloc (sizeof (*data));
	assert (data != NULL);

	/* Get all tokens until a non-number is found 
	   (it should be a command name) */
	data->argv[0] = strtok_r ((char *)command, DELIM, &save);
	if (my_atoi (data->argv[0]) == INT_MIN) {
		data->argv[1] = data->argv[0];
		data->argv[0] = (default_tid[0] == '\0') ? NULL: default_tid;
	}
	else
		data->argv[1] = strtok_r (NULL, DELIM, &save);

	/* At this point, both "tid" and "command" have been parsed, so
	 * let's check if the command is valid and take care of its
	 * arguments */
	
	for (data->argc = 1;  data->argv[data->argc] != NULL;
		 data->argv[++data->argc] = strtok_r (NULL, DELIM, &save))
		;

	/* argv[0] = <tid> or NULL if tid wasn't specified and there is no
	 *			 default value
	 * argv[1] = <command> */
	
    /* Is it a valid command? */
    cmdstruct = find_command (data->argv[1]);
	if (cmdstruct == NULL  ||  cmdstruct->cmd_funct == 0) {
		console_error (CE_INVALID_COMMAND);
		return;
	}

	/* Break the rest of the line into tokens */
	for (;
		 data->argc < MAX_ARGV  && data->argv[data->argc] != NULL;
		 data->argv[++data->argc] = strtok_r (NULL, cmdstruct->cmd_delim, &save))
		;

	assert (data->argv[data->argc] == NULL);

	/* Check for background commands */
	data->background = 0;
	lastarg = data->argv[data->argc-1];
	if (lastarg[0] == '&') {
		data->background = 1;
		data->argv[--data->argc] = NULL;
	}
	else if (lastarg[strlen (lastarg) - 1] == '&') {
		data->background = 1;
		((char *)lastarg)[strlen (lastarg) - 1] = '\0';
	}
	
	/* Execute the command */
	data->funct = cmdstruct->cmd_funct;
	execute_command_function (data);
}



PRIVATE RETSIGTYPE process_signals (int sig)
{
	printf ("\n==> Signal %d received\n", sig);
	execute_command ("quit");
    exit (0);
}


PRIVATE void logo (const char *level)
{
	int i;
	char package[64];

	strcpy (package, PACKAGE);
	for (i = 0;  i < strlen (package);  i++)
		package[i] = toupper (package[i]);
	
    puts ("\n");
    puts ("+---------------------------------------------+");
    printf ("|   %-6s - v%-10s   (Level %2s console) |\n",
			package, VERSION, level);
    puts ("+---------------------------------------------+");
    puts ("|   Copyright (c) 1999 DI - FCT/UNL           |");
    puts ("|                                             |");
    puts ("|   Type 'help' on prompt to get help         |");
    puts ("|   <TAB> expands commands                    |");
    puts ("|   Abbreviations are accepted if unambiguous |");
    puts ("+---------------------------------------------+");
}



PRIVATE int execute_command (const char *oline)
{
	char *line = strdup (oline);
	
	assert (line != NULL);
	execute (line);
	return 0;
}


PRIVATE int load_file (chain_t *file_list)
{
	char line[256];
	char *file;
	while ((file = chain_dequeue (file_list)) != NULL) {
		sprintf (line, "file %s", file);
		execute_command (line);
	}
	return 0;
}


PRIVATE int load_symbols (chain_t *file_list)
{
	char line[256];
	char *file = chain_dequeue (file_list);
	assert (file != NULL);
	sprintf (line, "symbol-file %s", file);
	execute_command (line);
	return 0;
}


PRIVATE int attach_process (int pid)
{
	char line[256];
	sprintf (line, "attach %d", pid);
	execute_command (line);
	return 0;
}


PRIVATE void redirect_channel (int channel, const char *device)
{
	int fd = open (device, channel == 0 ? O_RDONLY : O_WRONLY);
	assert (fd >= 0);
	dup2 (fd, channel);
}





PRIVATE int check_options (int argc, char *argv[])
{
	int ch;
	options.time = 0;
	options.input = NULL;
	options.output = NULL;
	options.file = chain_create();
	options.pid = 0;
#ifdef f2m
	options.server = NULL;
	options.port = NULL;
	options.flags = 0;
#define VALID_OPTIONS "h?ti:o:rks:p:a:x:"
#else
#define VALID_OPTIONS "h?ti:o:a:x:"
#endif
	while ((ch = getopt (argc, argv, VALID_OPTIONS)) != EOF) {
/* 		assert (ch != '?'); */
		switch (ch) {
		case 'h':
		case '?':
			usage (argv[0]); assert (0);
		case 't':
			options.time = 1; break;
		case 'i':
			assert (optarg != NULL); options.input = optarg; break;
		case 'o':
			assert (optarg != NULL); options.output = optarg; break;
#ifdef f2m
		case 'r':
			options.flags |= START_SERVER; break;
		case 'k':
			options.flags |= KEEP_SERVER; break;
		case 's':
			assert (optarg != NULL); options.server = optarg; break;
		case 'p':
			assert (optarg != NULL); options.port = optarg; break;
#endif
		case 'a':
			assert (optarg != NULL);
			options.pid = atoi(optarg);
			assert (options.pid > 0);
			break;
        case 'x':
			setenv ("FIDDLE_XDR", optarg, 1);
            break;
		}
	}
	for ( ;  optind < argc;  optind++)
		chain_enqueue(options.file, argv[optind]);
	assert (options.pid == 0  ||  chain_length (options.file) == 1) ;

	return optind;
}




int main (int argc, char *argv[])
{
	int status;
	char *fn;
	int level;

	/* Setup ^C to finish the console gracefully */
    signal (SIGINT, process_signals);
    signal (SIGTERM, process_signals);

	/* No default tid */
	strcpy (default_tid, "");

	/* Setup prompt */
	fn = strrchr (argv[0], '/');
	if (fn != NULL)
		++fn;
	else
		fn = argv[0];

	check_options (argc, argv);

	/* Process "-i" and "-o" command line options */
	if (options.input != NULL)
		redirect_channel (0, options.input);
	if (options.output != NULL)
		redirect_channel (1, options.output);
	
	/* Print logo */
	level = strcspn (fn, "0123456789");
	if (level == 0)
		logo ("  ");
	else
		logo (&fn[level]);

    /* Initialize readline */
#ifdef HAVE_LIBREADLINE
    rl_completion_entry_function = command_completion;
#endif

	/* Initialize the debugging library */
#ifdef f2m
	status = fx_initialize (options.server, options.port, options.flags);
#else
	status = fx_initialize ();
#endif
	if (status < 0)
		return status;

	if (options.pid != 0) {
		attach_process (options.pid);
		load_symbols (options.file);
	}
	else if (options.file != NULL)
		load_file (options.file);

	for (;;) {
		char *line = (char *)rl_gets (fn);
		if (line == NULL || strcmp (line, "quit") == 0)
			break;
/*  		nl = strdup (line); */
		execute_command (line);
	}
	execute_command (strdup ("quit"));
	return 0;
}
