/* *****************************************************************
 *
 *  $Id: chain.c,v 1.3 2002/09/27 12:43:03 jml Exp $
 *
 *  Copywright (c) 1999 by Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *  Time-stamp: <03/03/26 23:51:12 jml>
 *
 *  Universidade Nova de Lisboa
 *  Faculdade de Ciencias e Tecnologia
 *  Departamento de Informatica
 *  Parallel and Distributed Processing Group
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implements a double linked list
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "glob.h"
#include "chain.h"
#include "config.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#ifdef _REENTRANT
#include <pthread.h>

#define LOCK_C_MUX								\
		assert (&((chain_t *)chain)->c_mutex != NULL);		\
		pthread_mutex_lock (&((chain_t *)chain)->c_mutex)
#define UNLOCK_C_MUX							\
		assert (&((chain_t *)chain)->c_mutex != NULL);		\
		pthread_mutex_unlock (&((chain_t *)chain)->c_mutex)
#else
/* mutex opertations */
#define pthread_mutex_init(m,a)
#define pthread_mutex_destroy(m)
#define	pthread_mutex_lock(m)
#define	pthread_mutex_trylock(m)
#define	pthread_mutex_unlock(m)
/* condition opertations */
#define	pthread_cond_init(c,a)
#define	pthread_cond_destroy(c)
#define	pthread_cond_wait(c,m)
#define	pthread_cond_timedwait(c,m,t)
#define	pthread_cond_signal(c)
#define	pthread_cond_broadcast(c)
/* no lock nor unlock */
#define LOCK_C_MUX
#define UNLOCK_C_MUX
#endif

#define LOCK	LOCK_C_MUX
#define UNLOCK	UNLOCK_C_MUX




struct chain_collect_s {
	void	   *origin;
	void	   *data;
};


struct chain_node_s {
	void	   *data;
	chain_node_t     *next;
	chain_node_t     *prev;
};


struct chain_s {
	int     length;
	chain_node_t *first;
	chain_node_t *last;
#ifdef _REENTRANT
	pthread_mutex_t	mutex;
	pthread_mutex_t	c_mutex;
	pthread_cond_t	c_cond;
#endif
};



/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_node_create (void)
 *
 * Description: Create a new empty node
 *
 * Returns:     Pointer to new node
 */
PRIVATE chain_node_t *
chain_node_create (void)
{
	chain_node_t *node = calloc (1, sizeof (*node));
	assert (node != NULL);		/* Memory error */
	return node;
}


/* -----------------------------------------------------------------
 * Function:    void chain_node_destroy (chain_node_t *node)
 *
 * Description: Release the memory associated with a node
 *
 * Returns:     []
 */
PRIVATE void
chain_node_destroy (chain_node_t *node)
{
	assert (node != NULL);
	free (node);
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_append_nl (chain_t *chain, void *data)
 *
 * Description: Create a new node to keep 'data' and insert at in the 
 *						tail of the chain
 *				*** Assumes locks were acquired before ***
 *
 * Returns:     Pointer to new node
 */
PRIVATE chain_node_t *
chain_append_nl (chain_t *chain, void *data)
{
	chain_node_t *node;

	node = chain_node_create ();
	if (node != NULL) {
		node->data = data;
		node->next = NULL;
		node->prev = chain->last;
		
		if (chain->last == NULL)	/* First node in chain */
			chain->first = chain->last = node;
		else {						/* Chain not empty */
			chain->last->next = node;
			chain->last = node;
		}
		++chain->length;
		pthread_cond_broadcast (&(chain->c_cond));
	}
	return node;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_remove_nl (chain_t *chain, chain_node_t *node)
 *
 * Description: Remove a node from the chain.
 *				*** Assumes locks were acquired before ***
 *
 * Returns:     Pointer to node's data
 */
PRIVATE void *
chain_remove_nl (chain_t *chain, chain_node_t *node)
{
	void *data;

	if (chain->first == node)
		chain->first = node->next; /* We're removing the first node */
	else
		node->prev->next = node->next; /* It's not the first node */
	if (chain->last == node)
		chain->last = node->prev; /* We're removing the last node */
	else
		node->next->prev = node->prev; /* It's not the last node */
	data = node->data;
	chain_node_destroy (node);
	--chain->length;
	return (void *)data;
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_walk_nl (chain_t *chain,
 *						 		       chain_walk_f walker,
 *								       void *args)
 *
 * Description: Traverses the chain from head to tail applying the
 *				'walker()' function to the data object in each node.
 *				Stops immediately if 'wlaker()' returns negative.
 *				*** Assumes locks were acquired before ***
 *
 * Returns:     Pointer to the node where walking stopped, or NULL if
 *				all the chain was traversed.
 */
PRIVATE chain_node_t *
chain_walk_nl (chain_t *chain, chain_walker_f walker, void *args,
			   int stop_sign)
{
	chain_node_t *node = NULL;

	assert (stop_sign == 1  ||  stop_sign == -1);
	for (node = chain->first;  node != NULL;  node = node->next)
		if (walker (node->data, args) * stop_sign < 0)
/* 		if (walker (node->data, args) < 0) */
			break;	/* Stop immediately */
	return node;
}


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_walk_collect_nl (chain_t *chain,
 *										        chain_walk_collect_f walker,
 *										        void *args)
 *
 * Description: Traverses the chain from head to tail applying the
 *				'walker()' function to the data object in each node.
 *				Collects the pointers returned by 'walker()' in a chain.
 *				Stops immediately if 'wlaker()' returns NULL.
 *
 * Returns:     List with pointers returned by 'walker()'..
 */
PRIVATE chain_t *
chain_walk_collect_nl (chain_t *chain, chain_walker_collect_f walker, void *args)
{
	chain_t *results;
	chain_node_t *node;
	chain_collect_t *ret_node;
	void *ret;

	results = chain_create ();
	assert (results != NULL);
	for (node = chain->first;  node != NULL;  node = node->next) {
		ret = walker (node->data, args);
		if (ret == NULL)
			break;	/* Stop immediately */
		ret_node = malloc (sizeof (*ret));
		assert (ret_node != NULL);
		ret_node->origin = node->data;
		ret_node->data = ret;
		ret = chain_append_nl (results, ret_node);
		assert (ret != NULL);
	}
	return results;
}


/* -----------------------------------------------------------------
 * Function:    int add_element (void *obj, void *args)
 *
 * Description: Add 'obj' to (chain_t *)args;
 *
 * Returns:     
 */
PRIVATE int
add_element (void *obj, void *args)
{
	assert (args != NULL);
	chain_append_nl ((chain_t *)args, obj);
	return 0;
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    void *chain_data (const chain_node_t *node)
 *
 * Description: Get the data kept in the chain node
 *
 * Returns:     Pointer to the data in the node
 */
void *
chain_node_data (const chain_node_t *node)
{
	assert (node != NULL);
	return (void *)node->data;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_collect_origin (const chain_collect_t *node)
 *
 * Description: Get the origin kept in a collect node
 *
 * Returns:     Pointer to the origin in the collect node
 */
void *
chain_collect_origin (const chain_collect_t *node)
{
	assert (node != NULL);
	return (void *)node->origin;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_collect_data (const chain_collect_t *node)
 *
 * Description: Get the data kept in a collect node
 *
 * Returns:     Pointer to the data in the collect node
 */
void *
chain_collect_data (const chain_collect_t *node)
{
	assert (node != NULL);
	return (void *)node->data;
}


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_create (void)
 *
 * Description: Create a new empty chain
 *
 * Returns:     Pointer to new chain
 */
chain_t *
chain_create (void)
{
	chain_t *chain = malloc (sizeof (*chain));
	assert (chain != NULL);		/* Memory error */
	chain->first = chain->last = NULL;
	chain->length = 0;
	pthread_mutex_init (&(chain->c_mutex), pthread_mutexattr_default);
	pthread_cond_init (&(chain->c_cond), pthread_condattr_default);
	return chain;
}


/* -----------------------------------------------------------------
 * Function:    void chain_destroy (chain_t *chain)
 *
 * Description: Delete a chain.  The data in the nodes is "forgotten".
 *
 * Returns:     []
 */
void
chain_destroy (chain_t *chain)
{
	assert (chain != NULL);
	LOCK;
	while (chain->first != 0)
		chain_remove_nl (chain, chain->first);
	UNLOCK;
	pthread_mutex_destroy (&(chain->c_mutex));
	pthread_cond_destroy (&(chain->c_cond));
	free (chain);
}


/* -----------------------------------------------------------------
 * Function:    int chain_length (const chain_t *chain)
 *
 * Description: Get the number of nodes in chain
 *
 * Returns:     The number of nodes in chain
 */
int
chain_length (const chain_t *chain)
{
	int len;
	assert (chain != NULL);
	LOCK;
	len = chain->length;
	UNLOCK;
	return len;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_first (const chain_t *chain)
 *
 * Description: Get the first node of chain
 *
 * Returns:     Pointer to first node in chain
 */
void *
chain_first_all (const chain_t *chain, int value_flag)
{
	chain_node_t *node;
	
	assert (chain != NULL);
	LOCK;
	node = chain->first;
	UNLOCK;
	return (node == NULL  ||  value_flag == 0)
		? node : chain_node_data (node);
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_last (const chain_t *chain)
 *
 * Description: Get the last node of chain
 *
 * Returns:     Pointer to last node in chain
 */
void *
chain_last_all (const chain_t *chain, int value_flag)
{
	chain_node_t *node;
	
	assert (chain != NULL);
	LOCK;
	node = chain->last;
	UNLOCK;
	return (node == NULL  ||  value_flag == 0)
		? node : chain_node_data (node);
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_next (const chain_t *chain,
 *									const chain_node_t *node)
 *
 * Description: Get the successor of a node
 *
 * Returns:     Pointer to next node in chain
 */
chain_node_t *
chain_next (const chain_t *chain, const chain_node_t *node)
{
	chain_node_t *nnode;
	
	assert (chain != NULL);
	assert (node != NULL);
	LOCK;
	nnode = node->next;
	UNLOCK;
	return nnode;
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_prev (const chain_t *chain,
 *									const chain_node_t *node)
 *
 * Description: Get the predecessor of a node
 *
 * Returns:     Pointer to previous node in chain
 */
chain_node_t *
chain_prev (const chain_t *chain, const chain_node_t *node)
{
	chain_node_t *nnode;

	assert (chain != NULL);
	assert (node != NULL);
	LOCK;
	nnode = node->prev;
	UNLOCK;
	return nnode;
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_enqeue (chain_t *chain, void *data)
 *
 * Description: Create a new node to keep 'data' and insert at in the 
 *						tail of the chain
 *
 * Returns:     Pointer to new node
 */
chain_node_t *
chain_enqueue (chain_t *chain, void *data)
{
	chain_node_t *node;

	assert (chain != NULL);
	LOCK;
	node = chain_append_nl (chain, data);
	UNLOCK;
	return node;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_dequeue (chain_t *chain)
 *
 * Description: Remove the node at the heads of the chain
 *
 * Returns:     Pointer to node's data
 */
void *
chain_dequeue (chain_t *chain)
{
	void *data;

	assert (chain != NULL);
	LOCK;
	if (chain->first != NULL)
		data = chain_remove_nl (chain, chain->first);
	else
		data = NULL;
	UNLOCK;
	return data;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_dequeue_cond (chain_t *chain,
 *										  chain_walker_f *cond,
 *										  void *args)
 *
 * Description: Remove the first node from the chain for which the c_cond
 *				returns positive.
 *
 * Returns:     Pointer to node's data
 */
void *
chain_dequeue_cond (chain_t *chain, chain_walker_f cond, void *args)
{
	chain_node_t *node;
	void *data;

	LOCK;
	while ((node = chain_walk_nl (chain, cond, args, -1)) == NULL) 
		pthread_cond_wait (&(chain->c_cond), &(chain->c_mutex));
	data = chain_remove_nl (chain, node);
	UNLOCK;
	return data;
}


/* -----------------------------------------------------------------
 * Function:    void *chain_remove (chain_t *chain, chain_node_t *node)
 *
 * Description: Remove a node from the chain
 *
 * Returns:     Pointer to node's data
 */
void *
chain_remove (chain_t *chain, chain_node_t *node)
{
	void *data;

	assert (chain != NULL);
	assert (node != NULL);
	LOCK;
	data = chain_remove_nl (chain, node);
	UNLOCK;
	return data;
}


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_merge (chain_t *chain, chain_t *chain2)
 *
 * Description: Append a copy of all the elements in 'chain2' to 'chain'
 *
 * Returns:     Pointer to 'chain'
 */
chain_t *
chain_merge (chain_t *chain, chain_t *chain2)
{
	assert (chain != NULL);
	assert (chain2 != NULL);
	chain_walk (chain2, add_element, chain);
	return chain;
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_find (chain_t *chain,
 *									const void *what,
 *								    chain_cmp_f compare,
 *								    void *args)
 *
 * Description: Lookpuk for 'what' in chain, using 'compare()' to
 *						find equality.
 *				'compare()' should return the same as 'strcmp()'
 *
 * Returns:     Pointer to node with equal data, or NULL if not found.
 */
chain_node_t *
chain_find (chain_t *chain, const void *what, chain_compare_f compare, void *args)
{
	chain_node_t *node;

	assert (chain != NULL);
	assert (compare != NULL);
	LOCK;
	for (node = chain->first;  node != NULL;  node = node->next) {
		if (compare (node->data, what, args) == 0)
			break;				/* Found */
	}
	UNLOCK;
	return node;				/* 'node = NULL' if not found */
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_walk (chain_t *chain,
 *						 		    chain_walk_f walker,
 *								    void *args)
 *
 * Description: Traverses the chain from head to tail applying the
 *				'walker()' function to the data object in each node.
 *				Stops immediately if 'wlaker()' returns negative.
 *
 * Returns:     Pointer to the node where walking stopped, or NULL if
 *				all the chain was traversed.
 */
chain_node_t *
chain_walk (chain_t *chain, chain_walker_f walker, void *args)
{
	chain_node_t *node;
	
	assert (chain != NULL);
	assert (walker != NULL);
	LOCK;
	node = chain_walk_nl (chain, walker, args, 1);
	UNLOCK;
	return node;
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_rwalk (chain_t *chain,
 *								    chain_walk_f walker,
 *								    void *args)
 *
 * Description: Traverses the chain from tail to head applying the
 *						'walker()' function to the data object in
 *						each node.
 *				Stops immediately if 'walker()' returns negative.
 *
 * Returns:     Pointer to the node where walking stopped, or NULL if
 *				all the chain was traversed.
 */
chain_node_t *
chain_rwalk (chain_t *chain, chain_walker_f walker, void *args)
{
	chain_node_t *node;
	
	assert (chain != NULL);
	assert (walker != NULL);
	LOCK;
	for (node = chain->last;  node != NULL;  node = node->prev)
		if (walker (node->data, args) < 0)
			break;	/* Stop immediately */
	UNLOCK;
	return node;
}


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_prune (chain_t *chain,
 *								     chain_walk_f pruner,
 *								     void *args)
 *
 * Description: Remove all nodes from chain for which 'pruner()' 
 *						returns positive.
 *				Stops the pruning if 'pruner()' returns negative.
 *				If the data in the node was dynamically allocated, the
 *						'pruner()' can release that memory before returning.
 * Returns:     Pointer to the node where prunning stopped, or NULL if
 *				all the chain was traversed.
 */
chain_node_t *
chain_prune (chain_t *chain, chain_walker_f pruner, void *args)
{
	chain_node_t *node;
	chain_node_t *next;
	
	assert (chain != NULL);
	assert (pruner != NULL);
	LOCK;
	for (node = chain->first;  node != NULL;  node = next) {
		int val = pruner (node->data, args);
		if (val < 0) break;	/* Stop immediately */
		next = node->next;
		if (val > 0)
			chain_remove_nl (chain, node);
	}
	UNLOCK;
	return node;
}


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_walk_collect (chain_t *chain,
 *										     chain_walk_collect_f walker,
 *											 void *args)
 *
 * Description: Traverses the chain from head to tail applying the
 *				'walker()' function to the data object in each node.
 *				Collects the pointers returned by 'walker()' in a chain.
 *				Stops immediately if 'wlaker()' returns NULL.
 *
 * Returns:     List with pointers returned by 'walker()'..
 */
chain_t *
chain_walk_collect (chain_t *chain, chain_walker_collect_f walker, void *args)
{
	chain_t *results;
	
	assert (chain != NULL);
	assert (walker != NULL);
	LOCK;
	results = chain_walk_collect_nl (chain, walker, args);
	UNLOCK;
	return results;
}


/* -----------------------------------------------------------------
 * Function:    void * chain_walk_until_found(chain_t *chain,
 *										      chain_walker_f walker,
 *											  void *args)
 *
 * Description: Keeps walking until the walker function returns true
 *
 * Returns:     List with pointers returned by 'walker()'..
 */
void *
chain_walk_until_found(chain_t *chain, chain_walker_f walker, void *args)
{
	chain_node_t *result;
	chain_node_t *temp;
	void *data;
	
	assert(chain);
	   
	LOCK;
	result = chain_walk_nl(chain, walker, args, -1);
	
	while (result == NULL) {
		pthread_cond_wait (&(chain->c_cond), &(chain->c_mutex));
		result = chain_walk_nl(chain, walker, args, -1);
	}
	data = chain_remove_nl(chain, result);
	
	UNLOCK;
	return data;

}



/* ################################################################# */
/* ################################################################# */
/* For testing purposes only */
/* ################################################################# */
/* ################################################################# */
#ifdef TEST_MAIN

#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#include <math.h>

#ifdef _REENTRANT
#define MAX_NUMBER 1024

struct x {
	int tid;
	int num;
};

struct y {
	chain_t *chain;
	int     tid;
};

static void *enqueue_numbers (void *info)
{
	int i;
	struct x *buffer;
	chain_t *ch = ((struct y *)info)->chain;
	int      t = ((struct y *)info)->tid;
	
	for (i = 0;  i < MAX_NUMBER;  i++) {
		buffer = malloc (sizeof (*buffer));
		buffer->tid = t;
		buffer->num = i;
		chain_enqueue (ch, buffer);
	}
	return NULL;
}



static int cond_f (void *data, void *args)
{
	struct x *num = (struct x *)data;
	int tid = (int) args;

	if (num == NULL) {
		printf ("%s:%d:cond_f: Assertion `num != ((void *)0)' for tid %d failed.\n", __FILE__, __LINE__, tid);
	}
	if (num->tid == tid)
		return -1;
/* 	printf ("Falhou... %d\n", tid); */
	return 0;
}

static void *dequeue_numbers (void *info)
{
	int i;
	struct x *buffer;
	chain_t *ch = ((struct y *)info)->chain;
	int      t = ((struct y *)info)->tid;

	for (i = 0;  i < MAX_NUMBER;  i++) {
		buffer = chain_dequeue_cond (ch, cond_f, (void *)t);
		assert (buffer != NULL);
		assert (t == buffer->tid);
/* 		printf ("%3d  %d\n", t, buffer->num); */
	}
	return NULL;
}

static int test_enq_deq (int nthreads)
{
	int i;
	struct y *info;
	struct x *num;
	pthread_t *th = alloca (nthreads * sizeof (*th));
	chain_t *chain = chain_create ();
	assert (chain != NULL);
	for (i = 0;  i < nthreads;  i++) {
		info = alloca (sizeof (*info));
		info->chain = chain;
		info->tid = i;
		pthread_create (&th[i], pthread_attr_default, enqueue_numbers, info);
	}
	for (i = 0;  i < nthreads;  i++)
		pthread_join (th[i], NULL);
	for (i = 0;  i < nthreads * MAX_NUMBER;  i++) {
		num = (struct x *)chain_dequeue (chain);
		assert (num != NULL);
		printf ("%3d  %d\n", num->tid, num->num);
	}
	return 0;
}


static int test_enq_deq_cond (int nthreads)
{
	int i;
	struct y *info;
	struct x *num;
	pthread_t *th = alloca (nthreads * 2 * sizeof (*th));
	chain_t *chain = chain_create ();
	assert (chain != NULL);
/* 	puts ("Creating threads to enqueue numbers..."); */
	for (i = 0;  i < nthreads;  i++) {
		info = alloca (sizeof (*info));
		info->chain = chain;
		info->tid = i;
		pthread_create (&th[i*2], pthread_attr_default, enqueue_numbers, info);
	}
/* 	puts ("Creating threads to dequeue numbers..."); */
	for (i = 0;  i < nthreads;  i++) {
		info = alloca (sizeof (*info));
		info->chain = chain;
		info->tid = i;
		pthread_create (&th[i*2+1], pthread_attr_default, dequeue_numbers, info);
	}
/* 	puts ("Waitting for children..."); */
	for (i = 0;  i < 2 * nthreads;  i++)
		pthread_join (th[i], NULL);
	return 0;
}

#else  /* _REENTRANT */

static int compare_str (void *data, const void *what, void *args)
{
	return (strcmp (data, what) == 0);
}

static int prune_str (void *data, void *args)
{
	if (strcmp (data, args) == 0) {
		free ((void *)data);
		return 1;
	}
	return 0;
}

static int print_str (void *data, void *args)
{
	printf ("String = [%s]\n", (char *)data);
	return 1;
}

static void seq_test (void)
{
	chain_node_t *n;
	chain_t *chain;
	char *data1 = strdup ("Ola 1");
	char *data2 = strdup ("Ola 2");
	char *data3 = strdup ("Ola 3");

	chain = chain_create ();
	chain_enqueue (chain, data1);
	chain_enqueue (chain, data2);
	chain_enqueue (chain, data3);
	chain_walk (chain, print_str, NULL);
	puts ("");
	chain_prune (chain, prune_str, "Ola 2");
	chain_walk (chain, print_str, NULL);
	chain_destroy (chain);
}
#endif /* _REENTRANT */


main (int argc, char **argv)
{
#ifdef _REENTRANT
	if (argc != 2) {
		printf ("Usage: %s <n_threads>\n", argv[0]);
		exit (1);
	}
	test_enq_deq_cond (atoi (argv[1]));
#else
	seq_test ();
#endif
}
#endif /* TEST_MA*/
